#include "ground_truth.hpp"

using namespace std;
/*
    The routines provided here are used to calculate image positions from real coordinates,
    and real coordinates from image position.

    The calculations are not accurate enough to pinpoint the exact locations of items in the scene (albeit not that far off),
    but they are more than good enough to use for mapping out the parts of the image outside of the playing field,
    as well as mapping out the opponent robot.
    It also works well for figuring out if an object should be visible or not in the image.

    More accurate pinpointing might be achieved by better rectification and better tuning of parameters.

    The tuning parameters are:
    CCD size (horizontal, vertical)
    CCD resolution (horizontal, vertical)
    Focal Length
    View Height
    Object Height (in this case, corn height)
    Downwards angle

    The known parameters are: CCD size, CCD resolution and Object Height.
    The Focal Length is somewhat known, but the readings are skewed due to rectification.
    The View Height can be measured, but it's hard to know exactly where the *real* height is (due to optics etc)
    The Downwards angle is known at 23 degrees, but due to inaccuracies in mounting, slight movements in the plate it is mounted on,
    as well as the retification skewing things, it might not be correct.

    The tuning is set in the constructor.
    The values can be changed in real time using y,h,u,j,i,k in the GUI.

*/

double angle_from_position_and_orientation(int y, int x, double orientation) {
   double angle = (atan(y/(double) x))*180.0/M_PI; // center should be 0, left is negative, right is positive.
    if(y>=0 && x>=0) {
        // angle is between 0 and 90 deg.
    } else if(y>=0 && x<0) {
        angle+=180.0;
        // angle is between 90 and 180 deg.
    } else if(y<0 && x<0) {
        angle+=180.0;
        // angle is between 180 and 270 deg.
    } else {
        //angle+=360.0;
        // angle is between 270 deg and 0.
    }
    angle-=orientation*180.0/M_PI;
    if(angle >= 180.0) {
        angle-=360.0;
    } else if (angle <= -180.0) {
        angle+=360.0;
    }
    return angle;
}

CvPoint position_from_distance_and_orientation(double orientation, double y, double x) {
    CvPoint p;
    double rel_distance = sqrt(y*y+x*x);
    //printf("rel_distance: %f\n",rel_distance);
    double angle = acos(y/rel_distance);
    if(x < 0) {
        angle *=-1.0;
    }

    angle +=orientation;
    angle *=180.0/M_PI;
    if(angle >= 360.0) {
        angle-=360.0;
    } else if (angle < 0.0) {
        angle+=360.0;
    }
    if(angle >= 0 && angle < 90) {
        // x pos, y pos.
        p.y = sin(angle*M_PI/180.0)*rel_distance;
        p.x = cos(angle*M_PI/180.0)*rel_distance;
    } else if(angle >= 90.0 && angle < 180) {
        p.y = cos((angle-90.0)*M_PI/180.0)*rel_distance;
        p.x = -sin((angle-90.0)*M_PI/180.0)*rel_distance;
    } else if(angle >= 180.0 && angle < 270) {
        p.y = -sin((angle-180.0)*M_PI/180.0)*rel_distance;
        p.x = -cos((angle-180.0)*M_PI/180.0)*rel_distance;
    } else {
        p.y = -cos((angle-270.0)*M_PI/180.0)*rel_distance;
        p.x = sin((angle-270.0)*M_PI/180.0)*rel_distance;
    }
    printf("orientation: %f, angle: %f, Y: %f, X: %f, calculated: %d, %d \n",orientation*180.0/M_PI,angle, y, x, p.y,p.x);
    return p;
}


ground_truth::ground_truth()
{
    ccd.horizontal = 4.76; // mm (4.65*1024/1000 = 4.76)
    ccd.vertical = 3.571; // mm (4.65*768/1000 = 3.571)
    ccd.diagonal = 5.952; // mm (5.952 mm effective)
    ccd.res_vertical = 768;
    ccd.res_horizontal = 1024;
    focal_length = 2.88; // Tamron 3-8mm (real: 3.43. 2.78)
    view_height = 0.30;
    corn_height = 0.15;
    downwards_angle = 25.5; // should be 23.0, but the computer is tilted slightly, so this is not accurate.
}
double ground_truth::field_of_view_CCD(double sensor_size, double focal_length)
{
    return atan((sensor_size/2.0)/focal_length)*2.0*180.0/M_PI;
}

void ground_truth::setFL(double value) {
    focal_length = value;
}

double ground_truth::getFL() {
    return focal_length;
}

void ground_truth::setViewHeight(double meters) {
    view_height = meters;
}

double ground_truth::getViewHeight() {
    return view_height;
}

double ground_truth::getDownwardsAngle() {
    return downwards_angle;
}

void ground_truth::setDownwardsAngle(double value) {
    downwards_angle = value;
}

double ground_truth::getFOV_diagonal()
{
    return field_of_view_CCD(ccd.diagonal,focal_length);
}

double ground_truth::getFOV_horizontal()
{
    return field_of_view_CCD(ccd.horizontal,focal_length);
}

double ground_truth::getFOV_vertical()
{
    return field_of_view_CCD(ccd.vertical,focal_length);
}

double ground_truth::distance_from_vertical_angle(double angle, double height)
{
    return tan(angle*M_PI/180.0)*(view_height-height);
}

double ground_truth::distance_from_vertical_angle(double angle)
{
    return distance_from_vertical_angle(angle,0);
}

double ground_truth::vertical_angle_from_distance(double distance, double height)
{
    //distance in meters, angle in radians.
    return atan(distance/(view_height-height))*180.0/M_PI;
}

double ground_truth::vertical_angle_from_distance(double distance) {
    //distance in mm, angle in radians.
    return vertical_angle_from_distance(distance,0);
}


coord ground_truth::position_from_ground_pixel(int y, int x)
{
    coord p;
    p.y=(unsigned int) round(distance_from_vertical_index(y)*10.0);
    p.x=(int) round(horizontal_position_from_distance(distance_from_vertical_index(y),x)*10.0);
    return p;
}

double ground_truth::horizontal_position_from_distance(double distance, int x)
{
    return (x-(ccd.res_horizontal)/2.0)*pixel_size_at_distance(distance);
}

int ground_truth::horizontal_index_from_angle(double angle) {
    // printf("angle: %lf\t FOV: %lf\n",angle,getFOV_horizontal());
    return ccd.res_horizontal/2.0 + (angle/(getFOV_horizontal()))*ccd.res_horizontal+10; // adjustment for center.
}


double ground_truth::distance_from_vertical_index(int vertical_index)
{
    // calculate vertical angle from index:
    double view_angle_bottom = 90-downwards_angle-(getFOV_vertical()/2.0);
    double view_angle_top = 90-downwards_angle+(getFOV_vertical()/2.0);

    // Here I assume there is no spherical aberration (highly unlikely even with rectification)
    double angle = view_angle_top-(view_angle_top-view_angle_bottom)*vertical_index/(ccd.res_vertical-1);
    return distance_from_vertical_angle(angle);
}

double ground_truth::error_from_vertical_index(int vertical_index)
{
    return abs(distance_from_vertical_index(vertical_index+1)-distance_from_vertical_index(vertical_index));
}

double ground_truth::pixel_size_at_distance(double distance)
{
    return 1000.0*focal_length*1024/((distance*1000.0)*ccd.horizontal);
}
int ground_truth::ccd_coverage_in_pixels(double distance, double size)
{
    return size*pixel_size_at_distance(distance);
}

int ground_truth::ball_size_from_distance(double distance)
{
    return ccd_coverage_in_pixels(distance,0.10); // ball diameter = 10 cm = 0.1 m
}

CvPoint ground_truth::image_position_from_real_coord(int y, int x, double orientation, int z) {
    //double distance = sqrt(y*y+x*x);
    //double distance=y;
    double angle=angle_from_position_and_orientation(y,x,orientation);
    double distance = cos(angle*M_PI/180.0)*sqrt(y*y+x*x);
    /*
    if(y<0) {
        if(x<0) {
            angle+=180.0;
        } else {
            angle-=180.0;
        }
    }
    */
    //printf("REAL_COORD Y: %d X: %d angle: %f orientation: %f object: %f\n", y, x, angle, orientation, atan(x/(double) -y));
    return image_position_from_real(distance, angle, z);
}

CvPoint ground_truth::image_position_from_real(double distance, double angle, int height)
{
    // position must be given relative to ground and in mm (anything else is horribly messy anyway).
    CvPoint p;
    double view_angle_bottom = 90-downwards_angle-(getFOV_vertical()/2.0);
    double view_angle_top = 90-downwards_angle+(getFOV_vertical()/2.0);
    double vertical_angle = vertical_angle_from_distance(distance/(double) 1000.0, height/(double) 1000.0);
    int vertical_index = (view_angle_top-vertical_angle)/(view_angle_top-view_angle_bottom)*(ccd.res_vertical-1);
    int horizontal_index = horizontal_index_from_angle(angle);
    p.y=(vertical_index+60); // 30 for right camera
    p.x=horizontal_index+15; // 80 for left?
    //printf("vertical angle: %lf, vertical index: %d\n",vertical_angle,vertical_index);
    //printf("horizontal index, %d\n", horizontal_index);
    return p;
}




#ifdef STANDALONE
int main()
{
    ground_truth test = ground_truth();
    cout << test.distance_from_vertical_index(0) << endl;
    cout << test.distance_from_vertical_index(767) << endl;
    /*
    for (int i=200; i<500; i++)
    {
        double distance = test.distance_from_vertical_index(i);
        printf("%d: %lf m (error: %4.2lf mm)\n",i,distance, test.error_from_vertical_index(i)*1000);
        printf("%d: a ball would be %d pixels\n",i,test.ball_size_from_distance(distance));
    }
    */
    test.image_position_from_real_coord(1000,1,0);
    test.image_position_from_real_coord(1000,2000,0);
    return 0;
}
#endif
