#include "hough.hpp"
#include <cstdio>
/*
    The code here is essentially dedicated to just Circle Detection,
    which entails various Hough Circle Transform approaches.

    None of the approaches here use color information,
    so the input is merely one channel black&white images.
*/

using namespace std;

// This routine attemps to implement Fast Finding and Fitting (described in my report, under Circle detection)
// It only seems to work well with synthetic input, but this might be due to some programming error in the code below.
// It should nevertheless be a decent starting point for a FFF implemetation.
void fast_finding_and_fitting(BwImage& input, BwImage& edge_image, BwImage& accumulator, IplImage* raw_output)
{
    //input needs to have 1 channel.
    CvSize imageSize = input.getSize();

    CvMemStorage* storage = cvCreateMemStorage(0);
    BwImageShort F_h = cvCreateImage( imageSize, 16, 1);
    BwImageShort F_hv = cvCreateImage( imageSize, 16, 1);
    // detect edges
    wrapper::find_edges(input, edge_image);
    RgbImage output(raw_output);

    cvSet(accumulator.raw(), cvScalar(0));

    int min_dist=20;

    // scan edge image from left to right and top to bottom
    for(int x=0; x<imageSize.width; x++) {
        for(int y_top=0; y_top<imageSize.height; y_top++) {
            if(edge_image[y_top][x] != 0) {
                for(int y_bottom=y_top+min_dist; y_bottom<imageSize.height; y_bottom++) {
                    if(edge_image[y_bottom][x] !=0) {
                        int u=round((y_top+y_bottom)/2.0);
                        accumulator[u][x]++;
                    }
                }
            }
        }
    }

    /*
    // for testing.
    for(int y=0; y<imageSize.height; y++) {
        for(int x=0; x<imageSize.width; x++) {
            if(edge_image[y][x] != 0) {
                output[y][x].r=255;
                //output[y][x].g=accumulator[y][x];
            } else {
                output[y][x].r=0;
            }
            if(accumulator[y][x] != 0) {
                output[y][x].g=255;
            } else {
                output[y][x].g=0;
            }
            accumulator[y][x]=0;
        }
    }
    */

    // conduct HT in accumulator space to find all possible vertical symmetrical axes
    CvSeq* lines = cvHoughLines2( accumulator.raw(), storage, CV_HOUGH_STANDARD, 1, 5*CV_PI/180, 100, 0, 0 );
    vector<int> vertical_symmetry;
    vertical_symmetry.assign(imageSize.height,0);
    for(int i=0; i < lines->total; i++) {
        float* line = (float*) cvGetSeqElem(lines, i);
        float theta = line[1];
        if(theta >= 89.9*CV_PI/180.0 && theta <= 90.1*CV_PI/180.0) {
            float rho = line[0];
            float pos_y = rho;

            cout << "rho: " << pos_y << endl;
            vertical_symmetry[round(line[0])]++;
        }
    }
    cvSet(accumulator.raw(), cvScalar(0));

    // checking vertical symmetry.
    cout << "checking vertical symmetry" << endl;
    for(int x=0; x<imageSize.width; x++) {
        for(int y_top=0; y_top<imageSize.height; y_top++) {
            if(edge_image[y_top][x] != 0) {
                for(int y_bottom=y_top+1; y_bottom<imageSize.height; y_bottom++) {
                    if(edge_image[y_bottom][x] !=0) {
                        int u=round((y_top+y_bottom)/2.0);
                        if(vertical_symmetry[u] != 0) {
                            F_h[y_top][x]++;
                            F_h[y_bottom][x]++;
                        }
                    }
                }
            }
        }
    }

    // scan edge image F_h top to bottom, left to right.
    for(int y=0; y<imageSize.height; y++) {
        for(int x_left=0; x_left<imageSize.width; x_left++) {
            if(F_h[y][x_left] != 0) {
                for(int x_right=x_left+min_dist; x_right<imageSize.width; x_right++) {
                    if(F_h[y][x_right] !=0) {
                        int v=round((x_left+x_right)/2.0);
                        accumulator[y][v]++;
                    }
                }
            }
        }
    }

    // conduct HT in accumulator space to find all possible horizontal symmetrical axes
    CvSeq* lines2 = cvHoughLines2( accumulator.raw(), storage, CV_HOUGH_STANDARD, 1, 5*CV_PI/180, 100, 0, 0 );
    vector<int> horizontal_symmetry;
    horizontal_symmetry.assign(imageSize.width,0);
    for(int i=0; i < lines2->total; i++) {
        float* line = (float*) cvGetSeqElem(lines2, i);
        float theta = line[1];
        if(theta <= 0.1*CV_PI/180.0) {
            float rho = line[0];
            float pos_y = rho;
            cout << "rho: " << pos_y << endl;
            horizontal_symmetry[line[0]]++;
        }
    }
    cvSet(accumulator.raw(), cvScalar(0));

    // checking horizontal symmetry.
    cout << "checking horizontal symmetry" << endl;
   for(int y=0; y<imageSize.height; y++) {
        for(int x_left=0; x_left<imageSize.width; x_left++) {
            if(F_h[y][x_left] != 0) {
                for(int x_right=x_left+1; x_right<imageSize.width; x_right++) {
                    if(F_h[y][x_right] !=0) {
                        int v=round((x_left+x_right)/2.0);
                        if(horizontal_symmetry[v] != 0) {
                            F_hv[y][x_left]++;
                            F_hv[y][x_right]++;
                        }
                    }
                }
            }
        }
    }

    for(int y=0; y<imageSize.height; y++) {
        for(int x=0; x<imageSize.width; x++) {
            if(F_hv[y][x] != 0) {
                printf("%d %d: %d\n",y,x,F_hv[y][x]);
                output[y][x].r = 255;
            } else {
                output[y][x].r = 0;
            }
        }
    }
    cvReleaseMemStorage( &storage );
}

// Some of the code in "HoughOptimized" is from OpenCV, but it has been altered, and comments have been added.
// Also, the original OpenCV version crashes for some unknown reason.
// It fails with all the optimizations on ICC enabled, so there might be room for improvements performance wise.
#define hough_cmp_gt(l1,l2) (aux[l1] > aux[l2])
static CV_IMPLEMENT_QSORT_EX( icvHoughSortDescent32s, int, hough_cmp_gt, const int* )

#pragma optimize ( "", off)
void HoughOptimized(IplImage *raw_image, float dp, float min_dist,
                    int min_radius, int max_radius,
                    int canny_threshold, int acc_threshold,
                    CvSeq* circles, int circles_max)
{
    #pragma optimize ( "", on)
    const int SHIFT = 10, ONE = 1 << SHIFT, R_THRESH = 30;

    CvMat *dx = 0, *dy = 0;
    CvMat *edges = 0;
    CvMat *accum = 0;
    int* sort_buf = 0;
    CvMat* dist_buf = 0;
    CvMemStorage* storage = 0;

    int x, y, i, j, center_count, nz_count;
    int rows, cols, arows, acols;
    int astep, *adata;
    float* ddata;
    CvSeq *nz, *centers;
    float idp, dr;
    CvSeqReader reader;
    // Canny edge detector to get edges.
    edges = cvCreateMat( raw_image->height, raw_image->width, CV_8UC1 );
    cvCanny( raw_image, edges, MAX(canny_threshold/2,1), canny_threshold, 3 );

    // using Sobel-operator to get gradient in x and y direction.
    dx = cvCreateMat( raw_image->height, raw_image->width, CV_16SC1 );
    dy = cvCreateMat( raw_image->height, raw_image->width, CV_16SC1 );
    cvSobel( raw_image, dx, 1, 0, 3 );
    cvSobel( raw_image, dy, 0, 1, 3 );

    if ( dp < 1.f )
        dp = 1.f;
    idp = 1.f/dp;

    accum = cvCreateMat( cvCeil(raw_image->height*idp)+2, cvCeil(raw_image->width*idp)+2, CV_32SC1 );
    cvZero(accum);

    storage = cvCreateMemStorage();
    nz = cvCreateSeq( CV_32SC2, sizeof(CvSeq), sizeof(CvPoint), storage );
    centers = cvCreateSeq( CV_32SC1, sizeof(CvSeq), sizeof(int), storage );
    rows = raw_image->height;
    cols = raw_image->width;
    arows = accum->rows - 2;
    acols = accum->cols - 2;
    adata = accum->data.i;
    astep = accum->step/sizeof(adata[0]);

    // this calculates the gradient direction (returning values of +1,0,-1 for x and y)
    // and then it uses these to add points to potential matches.
    for ( y = 0; y < rows; y++ )
    {
        const uchar* edges_row = edges->data.ptr + y*edges->step;
        const short* dx_row = (const short*)(dx->data.ptr + y*dx->step);
        const short* dy_row = (const short*)(dy->data.ptr + y*dy->step);

        for ( x = 0; x < cols; x++ )
        {
            float vx, vy;
            int sx, sy, x0, y0, x1, y1, r, k;
            CvPoint pt;

            vx = dx_row[x];
            vy = dy_row[x];
            // continue for loop if the point isn't an edge.
            if ( !edges_row[x] || (vx == 0 && vy == 0) )
                continue;

            float mag = sqrt(vx*vx+vy*vy);
            assert( mag >= 1 );
            sx = cvRound((vx*idp)*ONE/mag);
            sy = cvRound((vy*idp)*ONE/mag);

            x0 = cvRound((x*idp)*ONE);
            y0 = cvRound((y*idp)*ONE);

            // follow gradient in positive and negative direction.
            // add them to an accumulator function.
            for ( k = 0; k < 2; k++ )
            {
                x1 = x0 + min_radius * sx;
                y1 = y0 + min_radius * sy;

                for ( r = min_radius; r <= max_radius; x1 += sx, y1 += sy, r++ )
                {
                    int x2 = x1 >> SHIFT, y2 = y1 >> SHIFT;
                    if ( (unsigned)x2 >= (unsigned)acols ||
                            (unsigned)y2 >= (unsigned)arows )
                        break;
                    adata[y2*astep + x2]++;
                }

                sx = -sx;
                sy = -sy;
            }

            pt.x = x;
            pt.y = y;
            cvSeqPush( nz, &pt );
        }
    }
    // count of non-zero pixels.
    nz_count = nz->total;
    if ( !nz_count )
        goto release;

    // find potential centers based on accumulation function, simply by looking at the accumulated
    // pixels and discarding pixels below a certain threshold,
    // and making sure the accumulated function is the largest in the 8-connected neighborhood.
    for ( y = 1; y < arows - 1; y++ )
    {
        for ( x = 1; x < acols - 1; x++ )
        {
            int base = y*(acols+2) + x;
            if ( adata[base] > acc_threshold &&
                    adata[base] > adata[base-1] && adata[base] > adata[base+1] &&
                    adata[base] > adata[base-acols-2] && adata[base] > adata[base+acols+2] )
                cvSeqPush(centers, &base);
        }
    }

    center_count = centers->total;
    if ( !center_count )
        goto release;

    sort_buf = (int*)cvAlloc( MAX(center_count,nz_count)*sizeof(sort_buf[0]) );
    cvCvtSeqToArray( centers, sort_buf );

    // sort centers from those with largest accumulator to smallest.
    icvHoughSortDescent32s( sort_buf, center_count, adata );
    cvClearSeq( centers );
    cvSeqPushMulti( centers, sort_buf, center_count );

    dist_buf = cvCreateMat( 1, nz_count, CV_32FC1 );
    ddata = dist_buf->data.fl;

    dr = dp;
    min_dist = MAX( min_dist, dp );
    min_dist *= min_dist;

    // find centers from most probable centers.
    for ( i = 0; i < centers->total; i++ )
    {
        int ofs = *(int*)cvGetSeqElem( centers, i );
        y = ofs/(acols+2) - 1;
        x = ofs - (y+1)*(acols+2) - 1;
        float cx = (float)(x*dp), cy = (float)(y*dp);
        int start_idx = nz_count - 1;
        float start_dist, dist_sum;
        float r_best = 0, c[3];
        int max_count = R_THRESH;

        for ( j = 0; j < circles->total; j++ )
        {
            float* c = (float*)cvGetSeqElem( circles, j );
            if ( (c[0] - cx)*(c[0] - cx) + (c[1] - cy)*(c[1] - cy) < min_dist )
                break;
        }

        if ( j < circles->total )
            continue;

        cvStartReadSeq( nz, &reader );
        for ( j = 0; j < nz_count; j++ )
        {
            CvPoint pt;
            float _dx, _dy;
            CV_READ_SEQ_ELEM( pt, reader );
            _dx = cx - pt.x;
            _dy = cy - pt.y;
            ddata[j] = _dx*_dx + _dy*_dy;
            sort_buf[j] = j;
        }

        cvPow( dist_buf, dist_buf, 0.5 );
        icvHoughSortDescent32s( sort_buf, nz_count, (int*)ddata );

        dist_sum = start_dist = ddata[sort_buf[nz_count-1]];
        for ( j = nz_count - 2; j >= 0; j-- )
        {
            float d = ddata[sort_buf[j]];

            if ( d > max_radius )
                break;

            if ( d - start_dist > dr )
            {
                float r_cur = ddata[sort_buf[(j + start_idx)/2]];
                if ( (start_idx - j)*r_best >= max_count*r_cur ||
                        (r_best < FLT_EPSILON && start_idx - j >= max_count) )
                {
                    r_best = r_cur;
                    max_count = start_idx - j;
                }
                start_dist = d;
                start_idx = j;
                dist_sum = 0;
            }
            dist_sum += d;
        }

        if ( max_count > R_THRESH )
        {
            c[0] = cx;
            c[1] = cy;
            c[2] = (float)r_best;
            cvSeqPush( circles, c );
            if ( circles->total >= circles_max )
                break;
        }
    }
    release:
    cvReleaseMat( &dist_buf );
    cvFree( &sort_buf );
    cvReleaseMemStorage( &storage );
    cvReleaseMat( &edges );
    cvReleaseMat( &dx );
    cvReleaseMat( &dy );
    cvReleaseMat( &accum );
}


// This is a basic, straight-forward implementation of the Hough Circle transform with fixed radius.
void HoughTransformCircle(BwImage& input, RgbImage& output, double radius)
{
    CvSize imageSize = input.getSize();
    vector< vector<int> > accumulator;
    vector<int> temp;
    temp.assign(imageSize.width,0);
    accumulator.assign(imageSize.height,temp);

    // for each pixel in image, accumulate for a certain radius
    // p.204 "Feature Extraction and Image Processing"
    int max_val=1;
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            if (input[y][x]!=0)
            {
                for (int deg_angle=0; deg_angle<360; deg_angle++)
                {
                    double t=(deg_angle*M_PI)/180.0;
                    int x0=round(x-radius*cos(t));
                    int y0=round(y-radius*sin(t));
                    if (x0<imageSize.width && x0>0 && y0<imageSize.height && y0>0)
                    {
                        accumulator[y0][x0]++;
                        if (accumulator[y0][x0]>max_val)
                        {
                            max_val=accumulator[y0][x0];
                        }
                    }

                }
            }
        }
    }
    cout << max_val << endl;
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            output[y][x].g=accumulator[y][x]*255/max_val;
        }
    }
}

typedef struct
{
    unsigned short radius;
    unsigned short value;
} ht_point;

// This is a basic, straight-forward implementation of the Hough Circle transform with a variable radius.
// It is extremely slow, especially for a large range of radiuses.
// It might have some use in non-realtime applications,
// or if it is known that the circle is within a small region of interest.
void HoughTransformCircle(BwImage& input, RgbImage& output, int min_radius, int max_radius)
{
    CvSize imageSize = input.getSize();
    vector< vector< vector<unsigned char> > > accumulator;
    vector< vector<unsigned char> > temp_2d;
    vector<unsigned char> temp;
    temp.assign(max_radius-min_radius,0);
    temp_2d.assign(imageSize.width,temp);
    accumulator.assign(imageSize.height,temp_2d);

    // for each pixel in image, accumulate for a certain radius
    // p.204 "Feature Extraction and Image Processing"
    int max_val=1;
    cerr << "Hough Transform:" << flush;
#pragma omp parallel for
    for (int y=0; y<imageSize.height; y++)
    {
        if (y % 10 == 0)
        {
            cerr << "." << flush;
        }
        for (int x=0; x<imageSize.width; x++)
        {
            if (input[y][x]!=0)
            {
                for (int radius=min_radius; radius<max_radius; radius++)
                {
                    for (int deg_angle=0; deg_angle<360; deg_angle++)
                    {
                        double t=(deg_angle*M_PI)/180.0;
                        int x0=round(x-radius*cos(t));
                        int y0=round(y-radius*sin(t));
                        if (x0<imageSize.width && x0>0 && y0<imageSize.height && y0>0)
                        {
                            accumulator[y0][x0][radius-min_radius]++;
                            if (accumulator[y0][x0][radius-min_radius]>max_val)
                            {
                                max_val=accumulator[y0][x0][radius-min_radius];
                            }
                        }
                    }
                }
            }
        }
    }
    cerr << max_val << endl;
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            int val=0;
            int radius_val=0;
            for (int radius=min_radius; radius<max_radius; radius++)
            {
                if (accumulator[y][x][radius-min_radius]>val)
                {
                    val=accumulator[y][x][radius-min_radius];
                    radius_val=radius;
                }
            }
            output[y][x].g=accumulator[y][x][radius_val-min_radius]*255/max_val;
            output[y][x].r=(radius_val-min_radius)*255/(max_radius-min_radius);
        }
    }
}

// this implementation uses Parameter decomposition (p.215)
void HoughTransformCircle(IplImage *raw_input, IplImage *raw_output)
{
    BwImage input(raw_input);
    RgbImage output(raw_output);
    CvSize imageSize = cvGetSize(raw_input);
    vector< vector<int> > accumulator;
    vector<int> temp;
    temp.assign(imageSize.width,0);
    accumulator.assign(imageSize.height,temp);

    // gather evidence
    int max_val=1;
#pragma omp parallel for
    for (int y1=1; y1<imageSize.height; y1++)
    {
        for (int x1=1; x1<imageSize.width; x1++)
        {
            if (input[y1][x1]!=0)
            {
                for (int y2=y1-12; y2<y1+12; y2++)
                {
                    for (int x2=x1-12; x2<x1+12; x2++)
                    {
                        if (abs(x2-x1)>10 || abs(y2-y1)>10)
                        {
                            if (x2>0 && y2>0 && x2<imageSize.width && y2<imageSize.height)
                            {
                                if (input[y2][x2]!=0)
                                {
                                    double xm=(x1+x2)/2.0;
                                    double ym=(y1+y2)/2.0;
                                    double m=99999999.0;
                                    if (abs(y2-y1)!=0)
                                    {
                                        m=(double) (x2-x1)/(y2-y1);
                                    }
                                    //cout << m << endl;
                                    if (m>-1.0 && m<1.0)
                                    {
                                        for (int x0=1; x0<imageSize.width; x0++)
                                        {
                                            int y0=round(ym+m*(xm-x0));
                                            if (y0>0 && y0<imageSize.height)
                                            {
                                                accumulator[y0][x0]++;
                                                if (accumulator[y0][x0]>max_val)
                                                {
                                                    max_val=accumulator[y0][x0];
                                                }
                                            }
                                        }
                                    }
                                    else
                                    {
                                        for (int y0=1; y0<imageSize.height; y0++)
                                        {
                                            int x0=round(xm+(ym-y0)/m);
                                            if (x0>0 && x0<imageSize.width)
                                            {
                                                accumulator[y0][x0]++;
                                                if (accumulator[y0][x0]>max_val)
                                                {
                                                    max_val=accumulator[y0][x0];
                                                }
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }
    }
    cout << max_val << endl;
    for (int y=0; y<imageSize.height; y++)
    {
        for (int x=0; x<imageSize.width; x++)
        {
            if (accumulator[y][x]==max_val)
            {
                cout << y << " " << x << endl;
            }
            output[y][x].g=accumulator[y][x]*255.0/max_val;
        }
    }
}

// implementing CHT w/gradient (naive implementation)
void Hough_Transform_Edge_Direction(frame<RgbPixel>& camera, RgbImage& output)
{
    // Edge extraction
    //cvCanny( raw_input, temp, 50, 200, 3);
    // Gradient images
    int vx_max=0;

    for(int y=0; y<camera.input.getHeight(); y++) {
        const short* dx_row = (const short*)(camera.dx->data.ptr + y*camera.dx->step);
        const short* dy_row = (const short*)(camera.dy->data.ptr + y*camera.dy->step);
        for(int x=0; x<camera.input.getWidth(); x++) {
            //if(edge_image[y][x] != 0) {
                float vx, vy, sx, sy;
                vx = dx_row[x];
                vy = dy_row[x];
                float mag = sqrt(vx*vx+vy*vy);
                if (mag < 1)
                {
                    continue;
                }
                //assert( mag >= 1 );
                //cout << "X: " <<  vx/mag << " Y: " << vy/mag << endl;
                sx = vx/mag;
                sy = vy/mag;
                //cout << "deg: " << sin(sy) << " " << cos(sx) << " " << tan(sx/sy) << endl;
                sin(sy);
                cos(sx);
                output[y][x].r=sqrt(vx*vx)/800*255;
                output[y][x].g=sqrt(vy*vy)/800*255;
                //output[y][x].r=(mag)*128+127;
                //output[y][x].r=sx*128.0+127;
                //output[y][x].g=sy*128.0+127;
            //}
        }
    }
}
