/*
Copyright (c) 2009-2010, Kai Hugo Hustoft Endresen.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the <organization> nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY Kai Hugo Hustoft Endresen ''AS IS'' AND ANY
EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <copyright holder> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

*/
#include "main.hpp"
#include "ground_truth.hpp"
#include "recognition/game_elements.hpp"
#include "board.hpp"
#include "communication.hpp"
#include "state.hpp"
#include <sstream>
#include <signal.h>
using namespace std;

// global variables, defined here to avoid multiple definitions linking error
struct timeval tp;
double zero;
int stop;
double time_left, time_right;
CvSize imageSize;
std::vector<int> global_ranges;
int global_rangesize;
CvPoint prev_pt, prev_rpt;
IplImage* display_output;
IplImage* histogram;
pthread_mutex_t show_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_t threads[3];

// added for histogram threshold.
void on_mouse( int event, int x, int y, int flags, void* param )
{
    if ( !exclusion )
        return;
    RgbImage color_match(exclusion);
    if ( event == CV_EVENT_LBUTTONUP ) //|| !(flags & CV_EVENT_FLAG_LBUTTON) )
        prev_pt = cvPoint(-1,-1);
    else if ( event == CV_EVENT_RBUTTONUP ) // || !(flags & CV_EVENT_FLAG_RBUTTON) )
        prev_rpt = cvPoint(-1,-1);
    else if ( event == CV_EVENT_RBUTTONDOWN )
        prev_rpt = cvPoint(x,y);
    else if ( event == CV_EVENT_LBUTTONDOWN )
        prev_pt = cvPoint(x,y);
    else if ( event == CV_EVENT_MOUSEMOVE && (flags & CV_EVENT_FLAG_LBUTTON) && !(flags & CV_EVENT_FLAG_RBUTTON) )
    {
        CvPoint pt = cvPoint(x,y);
        if ( prev_pt.x < 0 )
            prev_pt = pt;
        if (y<256)
        {
            cvLine( exclusion, prev_pt, pt, cvScalar(1,1,1), 5, 8, 0 );
        }
        else
        {
            intensity_exclusion[x]=255;
        }
        prev_pt = pt;
    }
    else if ( event == CV_EVENT_MOUSEMOVE && (flags & CV_EVENT_FLAG_RBUTTON) && !(flags & CV_EVENT_FLAG_LBUTTON) )
    {
        CvPoint pt = cvPoint(x,y);
        if (prev_rpt.x < 0)
            prev_rpt = pt;
        if (y<256)
        {
            cvLine( exclusion, prev_rpt, pt, cvScalar(0,0,0), 5, 8, 0);
        }
        else
        {
            intensity_exclusion[x]=0;
        }
        prev_rpt = pt;
    }
}

// reads a binary matrix from file.
void matrix_from_file(string filename, CvSize imageSize, CvMat* output)
{
    ifstream input(filename.c_str(), ios::binary);
    float* data = output->data.fl;
    for (int y=0; y< imageSize.height; y++)
    {
        for (int x=0; x < imageSize.width; x++)
        {
            input.read(reinterpret_cast<char*>(&data[y*imageSize.width+x]), sizeof(float));
        }
    }
    input.close();
}

void capture_to_file(IplImage* left_output, IplImage* right_output, string append)
{
    int timeval = time(0);
    string left,right;
    stringstream stream;
    stream << timeval << append << "_left" << ".png";
    stream >> left;
    stringstream stream2;
    stream2 << timeval << append << "_right" << ".png";
    stream2 >> right;
    cout << left << endl;
    wrapper::save_image(left_output,left);
    cout << right << endl;
    wrapper::save_image(right_output,right);
}
void capture_to_file(IplImage* left_output, IplImage* right_output) {
    capture_to_file(left_output, right_output, "");
}



// produces an RGB value for an intensity value.
void pixel_color(RgbPixel& pixel, int &element, int &rangesize, vector<int> &ranges, int &peak, int &bottom)
{
    if (peak<element) peak=element;
    if (bottom>element) bottom=element;
    pixel.r=0;
    pixel.g=0;
    pixel.b=0;
    if (element<ranges[0] && element>=ranges[1])
    {
        pixel.b=(element-ranges[1])*255.0/rangesize;
    }
    else if (element>=ranges[2])
    {
        pixel.g=(ranges[1]-element)*255.0/rangesize;
    }
    else if (element>=ranges[3])
    {
        pixel.g=(element-ranges[3])*255.0/rangesize;
    }
    else if (element>=ranges[4])
    {
        pixel.r=(ranges[3]-element)*255.0/rangesize;
    }
}

// produces a color-intensity image from an intensity value matrix.
void value_to_color(CvMat* raw_disp, IplImage* raw_output,int minval, int maxval)
{
    RgbImage output(raw_output);
    MatShort input(raw_disp);
    int element;
    vector<int> ranges;
    ranges.assign(4,0);
    int rangesize=(maxval-minval)/5.0;
    for (int i=0; i<5; i++)
    {
        ranges[i]=maxval-rangesize*i;
    }
    int peak=0, bottom=65536;
    int y=0, x=0;
    //#pragma omp parallel for
    for (y=0; y<raw_output->height; y++)
    {
        for (x=0; x<raw_output->width; x++)
        {
            element = (int) input[y][x];
            element+=32768;
            pixel_color(output[y][x], element, rangesize, ranges, peak, bottom);
        }
    }
    cout << "peak: " << peak << " bottom: " << bottom << endl;
}

// function to simplify getting time.
double get_time()
{
    gettimeofday(&tp,NULL);
    return static_cast<double>(tp.tv_sec) + static_cast<double>(tp.tv_usec)/1000000.0;
}

// calculates the distance to an object from the disparity.
double disparity_to_distance(int disparity)
{
    return 4750.0/disparity;
}

// calculates the amount of pixels in the x direction available at a certain distance, used to calculate position.
double distance_to_view_x(double distance)
{
    double angle=37.5; // 37.5 is default.
    return (distance/2.0)*tan(angle*M_PI/180.0)*2.0;
}

// calculates the amount of pixels in the y direction available at a certain distance, used to calculate position.
double distance_to_view_y(double distance)
{
    double angle=30.0;
    return (distance/2.0)*tan(angle*M_PI/180.0)*2.0;
}

void saveImage(IplImage *img_left, IplImage *img_right)
{
    int timeval = time(0);
    string left,right;
    stringstream stream;
    stream << timeval << "_left" << ".png";
    stream >> left;
    stringstream stream2;
    stream2 << timeval << "_right" << ".png";
    stream2 >> right;
    cout << left << endl;
    cvSaveImage(left.c_str(), img_left);
    cout << right << endl;
    cvSaveImage(right.c_str(), img_right);
}



void displayer()
{

}

void load_images_into_frames(char* left, char* right)
{
    options.input=1;
    RgbImage left_temp = cvLoadImage(left);
    RgbImage right_temp = cvLoadImage(right);
    if(options.color == 0) {
        // TODO: Fix BW
        //wrapper::convert_rgb2gray(left_temp,cam_left.input);
        //wrapper::convert_rgb2gray(right_temp,cam_right.input);
    } else {
        wrapper::copy_image(left_temp,cam_left.input);
        wrapper::copy_image(right_temp,cam_right.input);
    }
    //cam_left.ready=1;
    //cam_right.ready=1;
}

int thread_setup() {
    // barriers
    if (pthread_barrier_init(&barrier_before_processing, NULL, 2))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
    if (pthread_barrier_init(&barrier_after_processing, NULL, 3))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
    if (pthread_barrier_init(&barrier_correspondence, NULL, 3))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
    setGlobalState(STATE_WAITING);
    pthread_create(&threads[0], NULL, capture_thread, NULL);
    pthread_create(&threads[1], NULL, slave_thread, NULL);
    pthread_create(&threads[2], NULL, communication_thread, NULL);
}

int thread_cleanup() {
    int rc;
    // waiting for threads to finish
    printf("waiting for threads to end barrier\n");
    pthread_barrier_wait(&barrier_after_processing);
    printf("waiting for threads to finish..\n");
    for (int t=0; t<3; t++)
    {
        printf("thread %d...\n",t);
        rc = pthread_join(threads[t], NULL);
    }
    capture_cleanup();
}

int ignore_barriers() {
    if (pthread_barrier_init(&barrier_before_processing, NULL, 1))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
    if (pthread_barrier_init(&barrier_after_processing, NULL, 1))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
    if (pthread_barrier_init(&barrier_correspondence, NULL, 1))
    {
        printf("Could not create a barrier\n");
        return -1;
    }
}

void catch_int(int sig_num){
	signal(SIGINT, catch_int);
	cout << "Ctrl+C interrupted program" << endl;
    send_stop_to_self();
    stop=1;
	thread_cleanup();
}


int main (int argc, char **argv)
{
    cout << "Tracking objects in 3D using Stereo-Vision" << endl;
    cout << "(C) 2009-2010 Kai Hugo Hustoft Endresen. All Rights Reserved." << endl << endl;
    cvUseOptimized(1);
    signal(SIGINT, catch_int);
    cout << "Color conversion options:" << endl
    << "0: YUYV->BW" << endl
    << "1: YUYV->YUV" << endl
    << "2: YUYV->RGB" << endl
    << "3: YUYV->YUV (thresholded)" << endl
    << "Color[0-3]: " << endl;
    cin >> options.color;
    cout << "Algorithms/Options:" << endl
    << "0: capture" << endl
    << "1: anaglyphic stereo" << endl
    << "2: Canny edge detection with edge direction" << endl
    << "3: set colors for thresholding (and testing)" << endl
    << "4/5: hough matching correspondence" << endl
    << "6: Hough Circle (specific radius)" << endl
    << "7: Hough Circle (unknown radius) (SLOW!)" << endl
    << "8: Hough Circle (unknown radius, thresholded) (SLOW!)" << endl
    << "9: Block-matching correspondence (current implementation is slow, working poorly.)" << endl
    << "10: Canny edge detection with adjacent colors" << endl
    << "11: Edge Block matching" << endl
    << "12: Harris Corner detection" << endl
    << "13: Normalization" << endl
    << "14: Color Segmentation (R.E. Blake)" << endl
    << "15: Fast Finding and Fitting - Symmetrical shape tracker (Circles)" << endl
    << "16: Color recognition using HSV space + edges" << endl
    << "17: Color segmentation + hole filling + smoothing" << endl
    << "18: Color segmentation + smoothing (used to show segmentation results)" << endl
    << "19: Normalization and RGB Gradient" << endl
    << "20: Find cylinders" << endl
    << "21: Find game elements (SOLUTION)" << endl
    << "Option[0-21]: " << endl;
    cin >> options.option;
    //options.sync_error=10;
    //options.color=atoi(argv[1]);
    //options.option=atoi(argv[2]);
    if (argc>3)
    {
        options.sync_error=atoi(argv[3]);
    }
    else
    {
        options.sync_error=10;
    }
    int channels=3;
    if (options.color==0)
    {
        channels=1;
    }
// initialize matrices
    imageSize = cvSize(1024,768);
    init_matrices(cam_left,channels);
    init_matrices(cam_right,channels);
    RgbImage small_temp = cvCreateImage(cvSize(320,240),8,3);
    RgbImage small_output = cvCreateImage(cvSize(320,240),8,3);
    histogram = cvCreateImage(cvSize(256,320), 8, 3);
    IplImage *raw_output = cvCreateImage(imageSize, 8, 3);
    display_output = cvCreateImage(imageSize, 8, 3);

    //left_circles = 0;
    //right_circles = 0;

    exclusion = cvCreateImage(cvSize(256,320),8,3);
/*
    exclusion = cvLoadImage("exclusion.png"); //,exclusion,CV_RGB2GRAY);
    RgbImage test(exclusion);
    for (int y=0; y<320; y++)
    {
        for (int x=0; x<256; x++)
        {
            if (test[y][x].r==0)
            {
                test[y][x].r=255;
            }
        }
    }
*/
    //cvSaveImage("exclusion.png",exclusion);
    intensity_exclusion.assign(256,0);

    cam_left.mx = cvCreateMat(imageSize.height,imageSize.width,CV_32F);
    cam_left.my = cvCreateMat(imageSize.height,imageSize.width,CV_32F);
    cam_right.mx = cvCreateMat(imageSize.height,imageSize.width,CV_32F);
    cam_right.my = cvCreateMat(imageSize.height,imageSize.width,CV_32F);
    IplImage *gradient_output = cvCreateImage(imageSize, 16,1);

    matrix_from_file("mx1.bin",imageSize,cam_left.mx);
    matrix_from_file("mx2.bin",imageSize,cam_right.mx);
    matrix_from_file("my1.bin",imageSize,cam_left.my);
    matrix_from_file("my2.bin",imageSize,cam_right.my);

    //cvMemStorage *storage = cvMemStorage(0);



    stop=0;
    cout << fixed;
    int rc;
    char key;
    zero = get_time();
    if(argc<2) {
        rc=capture_setup(imageSize.width,imageSize.height,7);
        if(rc == -1) {
            return -1;
        }
        thread_setup();
    } else {
        printf("load images into frames..\n");
        load_images_into_frames(argv[1],argv[2]);
        ignore_barriers();
        compute_on_frame(cam_left,0.0,0.0);
        compute_on_frame(cam_right,0.0,0.0);
    }

    vector<int> left_lines;
    vector<int> right_lines;
    initialize_element();

// for colors
    global_ranges.assign(5,0);
    global_rangesize=(300-0)/4.0;
    for (int i=0; i<5; i++)
    {
        global_ranges[i]=300-global_rangesize*i;
    }
    int peak=0, bottom=65536;
    vector<unsigned char> pixels;
    BwImage accum = cvCreateImage( imageSize, 8, 1);
    BwImage temp = cvCreateImage( imageSize, 8, 1);
    BwImage temp2 = cvCreateImage( imageSize, 8, 1);
    IplImage* location = cvCreateImage( cvSize(800,600),8,3);
    //CvMemStorage* store = cvCreateMemStorage(0);
    //CvSeq* left_circles;
    //CvSeq* right_circles;
    double start = get_time();
    int matches=0;
    //sleep(5);
    cvNamedWindow("disparity");
    cvNamedWindow("left");
    //cvNamedWindow("edges");
    cvNamedWindow("right");
    cvNamedWindow("left_rectified");
    cvNamedWindow("right_rectified");
    //cvNamedWindow("location");
    if (options.option==3)
    {
        cvNamedWindow("histogram");
    }
    double old_left=0.0;
    double old_right=0.0;
    int counter=0;
    CvMat* harris_output = cvCreateMat(imageSize.height, imageSize.width, CV_32FC1);
    RgbImage left_input(cam_left.remapped);
    RgbImage left_output(cam_left.output);
    RgbImage output(raw_output);
    double color_time;
    int loops=0;
    CvSeq* comp = NULL;
    init_game_elements();
    if(argc < 2) { // no need set program to cylinder config state when only working with images.
        sleep(120);
        if(getGlobalState() == STATE_WAITING) {
            our_color=0;
            setGlobalState(STATE_CYLINDER_CONFIG);
        }
    }
    pthread_barrier_wait(&barrier_correspondence);
    vector< CvPoint2D32f > tomatoes;
    vector< CvPoint2D32f > oranges;
    vector< CvPoint2D32f > corns;
    while (!stop)
    {
        // RLE:
        vector< vector < run_len >  > rle_left;
        vector< vector < run_len >  > rle_right;
        vector < run_len > temp_run;
        rle_left.assign(imageSize.height,temp_run);
        rle_right.assign(imageSize.height,temp_run);
        //cout << "performing.." << endl;
        cvSet(temp.raw(),cvScalar(0));
        cvSet(histogram, cvScalar(0,0,0));
        cvSet(raw_output, cvScalar(0,0,0));
        pthread_barrier_wait(&barrier_after_processing);
        //cout << abs(cam_left.time-cam_right.time)*1000 << endl;
        //IplImage* left = (IplImage*) cvClone(left_remapped);
        //IplImage* right = (IplImage*) cvClone(right_remapped);
        //pthread_mutex_unlock(&left_mutex);
        //pthread_mutex_unlock(&right_mutex);
        switch (options.option)
        {
        case 0:
            if (counter % 5 == 0)
            {
		//anaglyphic_stereo(raw_output);
                capture_to_file(cam_left.remapped.raw(), cam_right.remapped.raw());
                //capture_to_file(cam_left.intensity.raw(), cam_right.intensity.raw(), "bw");
            }
            break;
        case 1:
            anaglyphic_stereo(raw_output);
            break;
        case 2:
            edge_data(cam_left.remapped.raw(),raw_output);
            //edge_data_verbose(cam_left,raw_output);
            //cvSaveImage("output_left.png",left_remapped);
            break;
        case 3:
            thresholding_test(cam_left.remapped.raw(),histogram);
            cvShowImage("rg_histogram",histogram);
            cvSetMouseCallback( "rg_histogram", on_mouse, 0 );
            break;
        case 4:
            if (cam_left.circles && cam_right.circles)
            {
                matches++;
                hough_matching_correspondence(cam_left.circles, cam_right.circles,raw_output,location);
            }
            break;
        case 5:
            //sleep(1);
            if (cam_left.circles && cam_right.circles)
            {
                matches++;
                hough_matching_correspondence(cam_left.circles, cam_right.circles,raw_output,location);
                for (int i=0; i < cam_left.circles->total; i++)
                {
                    float *left_p = (float *) cvGetSeqElem( cam_left.circles, i);
                    cout << "left: " << left_p[0] << " " << left_p[1] << " " << left_p[2] << "(out of " << cam_left.circles->total << ")" << endl;
                    cvCircle( cam_left.remapped.raw(), cvPoint(cvRound(left_p[0]),cvRound(left_p[1])), cvRound(left_p[2]), CV_RGB(255,0,0), 3, 8, 0);
                }
                for (int i=0; i < cam_right.circles->total; i++)
                {
                    float *right_p = (float *) cvGetSeqElem( cam_right.circles, i);
                    cout << "right: " << right_p[0] << " " << right_p[1] << " " << right_p[2] << "(out of " << cam_right.circles->total << ")" << endl;
                    cvCircle( cam_right.remapped.raw(), cvPoint(cvRound(right_p[0]),cvRound(right_p[1])), cvRound(right_p[2]), CV_RGB(255,0,0), 3, 8, 0);
                }
                cvShowImage("location",location);
            }
            break;
        case 6:
            wrapper::convert_rgb2gray(cam_left.remapped,cam_left.ch1);
            wrapper::find_edges(cam_left.ch1, temp);
            HoughTransformCircle(temp,output,130.0);
            //HoughTransformCircle(temp,raw_output);
            break;
        case 7:
            //thresholding_test(left_remapped, histogram);
            //YUVtoBW(left_remapped,temp,imageSize);
            // Todo: Fix BW
            //wrapper::find_edges(cam_left.remapped, temp2);
            HoughTransformCircle(temp2,output,20,options.sync_error);
            break;
        case 8:
            thresholding_test(cam_left.remapped.raw(), histogram);
            YUVtoBW(cam_left.remapped.raw(),temp.raw(),imageSize);
            wrapper::find_edges(temp, temp2);
            HoughTransformCircle(temp2.raw(),raw_output);
            break;
        case 9:
            block_matching_correspondence(cam_left.remapped.raw(), cam_right.remapped.raw(), raw_output);
            break;
        case 10:
            edge_data_test(cam_left,raw_output);
            break;
        case 11:
            edge_block_matching(raw_output);
            break;
        case 12:
            YUVtoBW(cam_left.remapped.raw(),temp.raw(),imageSize);
            cvCornerHarris(temp.raw(), harris_output, 4, 3, 0.1);
            wrapper::scale_8bit(harris_output, temp);
            //wrapper::convert_rgb2gray(temp,raw_output);
            break;
        case 13:
            for(int y=0; y<imageSize.height; y++) {
                for(int x=0; x<imageSize.width; x++) {
                    normalize(left_input[y][x],120.0);
                }
            }
            //fast_finding_and_fitting(cam_left.remapped, temp, temp2, raw_output);
            //Hough_Transform_Edge_Direction(cam_left, output);
            //color_segmentation(cam_left.remapped,raw_output,70.0,1);
            break;
        case 14:
        /*
            for(int y=0; y<imageSize.height; y++) {
                for(int x=0; x<imageSize.width; x++) {
                    normalize(cam_left.remapped[y][x],100.0);
                }
            }
            */
            cvSet(small_output.raw(), cvScalar(0,0,0));
            //cvResize(cam_left.remapped.raw(),small_temp.raw());
            color_time = get_time();

            color_segmentation(cam_right.remapped.raw(), output.raw(), 10.0,2);
            cout << (get_time()-color_time)*1000.0 << endl;
            break;
        case 15:
            if(options.color==0) {
                //BW_Gradient(cam_left.remapped, gradient_output, 2);
                //wrapper::gradient_horizontal(cam_left.remapped, cam_left.dx);
                //cvNormalize(cam_left.dx,temp,0,255,CV_MINMAX);
                value_to_color(cam_left.dx,raw_output,32768,36000);
            } else {
                wrapper::convert_rgb2gray(cam_left.remapped,temp);
                fast_finding_and_fitting(temp, temp2, accum, raw_output);
                /*
                double time_gradient=get_time();
                RGB_Gradient(cam_left.remapped.raw(), cam_left.dx, 2);
                cout << (get_time()-time_gradient)*1000.0 << endl;
                value_to_color(cam_left.dx,raw_output,32500, 33500);
                */
                //wrapper::scale_8bit(cam_left.dx,temp);
            }
            //cvCvtColor(temp.raw(),raw_output,CV_GRAY2BGR);
            break;
        case 16:
            detect_colors(cam_left.remapped, output);
            wrapper::convert_rgb2gray(cam_left.remapped, temp);
            wrapper::find_edges(temp,temp2);
            //RGB_Gradient(cam_left.remapped.raw(), cam_left.dx, 2);
            //wrapper::scale_8bit(cam_left.dx,temp);
            for(int y=0; y<output.getHeight(); y++) {
                for(int x=0; x<output.getWidth(); x++) {
                    if(temp2[y][x] != 0) {
                        output[y][x].r=255;
                        output[y][x].g=255;
                        output[y][x].b=255;
                    }
                }
            }
            break;
        case 17:
            color_time = get_time();
            for(int y=0; y<cam_left.output.getHeight(); y++) {
                for(int x=0; x<cam_left.output.getWidth(); x++) {
                    if(!pixel_is_zero(cam_left.diff[y][x])) {
                        cam_left.output[y][x]=cam_left.diff[y][x];
                    }
                    if(!pixel_is_zero(cam_right.diff[y][x])) {
                        cam_right.output[y][x]=cam_right.diff[y][x];
                    }
                }
            }
/*
            run_length_encode_diff(cam_left.output,rle_left);
            run_length_encode_diff(cam_right.output,rle_right);
            RLE_match(rle_left,rle_right,output,10,30);
            */
            /*
            for(int y=0; y<cam_left.output.getHeight(); y++) {
                for(int x=0; x<cam_left.output.getWidth(); x++) {
                    if(!pixel_is_zero(cam_left.diff[y][x])) {
                        output[y][x]=cam_left.diff[y][x];
                    } else {
                        output[y][x]=cam_left.output[y][x];
                    }
                }
            }
            */
            //region_fix(cam_left.output, cam_left.remapped, 100, 100);
            cout << (get_time()-color_time)*1000.0 << endl;

/*

        fill_holes(cam_left.output, output, 100);
        wrapper::convert_rgb2gray(output, temp);
        matches=-1;
        loops=0;
        while(matches != 0) {
            loops++;
            matches=skeletonize(temp);
        }
        cvCvtColor(temp.raw(),raw_output,CV_GRAY2BGR);
*/
            /*
            wrapper::convert_rgb2gray(output, temp);
            cvDistTransform( temp.raw(), temp2.raw(), CV_DIST_L1, 3);
            wrapper::scale_8bit(temp2,cam_left.temp);
            cvCvtColor(cam_left.temp.raw(),raw_output,CV_GRAY2BGR);
            */
//            cout << "SKELETONIZE: " << matches << " required " << loops << " loops." << endl;
  //          cout << (get_time()-color_time)*1000.0 << endl;

            //fill_holes(cam_left.output, output, 100);
            //cvDrawContours(output.raw(), comp, cvScalar(26,26,26), cvScalar(128,128,128), 2,3);
            //fill_holes(cam_left.output, output, 100);
        /*
            color_time = get_time();
            RGB_Gradient(cam_left.remapped.raw(), cam_left.dx, 2);
            color_encode(cam_left.remapped, cam_left.dx, output);
            //detect_colors(cam_left.diff, output);
            cout << (get_time()-color_time)*1000.0 << endl;
            */
            /*
            wrapper::convert_rgb2gray(cam_left.remapped, temp);
            wrapper::find_edges(temp,temp2);
            //RGB_Gradient(cam_left.remapped.raw(), cam_left.dx, 2);
            //wrapper::scale_8bit(cam_left.dx,temp);

            for(int y=0; y<output.getHeight(); y++) {
                for(int x=0; x<output.getWidth(); x++) {
                    if(temp2[y][x] != 0) {
                        output[y][x].r=255;
                        output[y][x].g=255;
                        output[y][x].b=255;
                    }
                }
            }
            */
            break;
        case 18:
        /*
            //detect_colors(cam_left.output, output);
            for(int y=0; y<cam_left.output.getHeight(); y++) {
                for(int x=0; x<cam_left.output.getWidth(); x++) {
                    if(pixel_is_zero(cam_left.output[y][x])) {
                        cam_left.ch1[y][x]=255;
                    } else {
                        cam_left.ch1[y][x]=0;
                    }
                }
            }
            cvDistTransform( cam_left.ch1.raw(), temp.raw(), CV_DIST_L1, 3);
            wrapper::scale_8bit(temp,cam_left.temp);
            cvCvtColor(cam_left.temp.raw(),raw_output,CV_GRAY2BGR);
            color_time = get_time();
            for(int distance=1; distance<4; distance++) {
                matches=0;
                for(int y=0; y<output.getHeight(); y++) {
                    for(int x=0; x<output.getWidth(); x++) {
                        if(temp[y][x]==distance) {
                            cam_left.output[y][x].b=255;
                            matches++;
                        }
                    }
                }
                cout << "distance: " << distance << " matches: " << matches << endl;
            }
            cout << (get_time()-color_time)*1000.0 << endl;
*/
        //fill_holes(cam_left.output,150);

        /*
        color_time = get_time();
            cvPyrSegmentation(cam_left.remapped.raw(), output.raw(), cam_left.storage, &comp, 3, 200, 100);
            cout << (get_time()-color_time)*1000.0 << endl;
            */
            break;
        case 19:
            wrapper::scale_8bit(cam_left.dx,temp);
            cvCvtColor(temp.raw(),raw_output,CV_GRAY2BGR);
            //detect_colors(cam_left.remapped, output);
            break;

        case 20:
            //detect_colors(cam_left.remapped,output);
            //wrapper::copy_image(cam_right.remapped,output);
            all_cylinders(cam_left.remapped, cam_right.remapped,cam_left.timestamp);
            /*
            cvLine(output.raw(), cvPoint(356,0), cvPoint(356,500),cvScalar(1,1,1), 5, 8, 0);
            cvLine(output.raw(),
            */
            break;
        case 21:
            //wrapper::scale_8bit(cam_left.dx,temp);
            //cvCvtColor(temp.raw(),raw_output,CV_GRAY2BGR);
            //border_draw(cam_left, cam_right, cam_left.timestamp);
            corresponding_objects(cam_left,cam_right,tomatoes,oranges,corns,output);
            //update_elements(cam_left.timestamp,tomatoes,oranges,corns);
            //tomatoes.clear();
            //oranges.clear();
            //corns.clear();
            break;
        case 22:
            all_cylinders(cam_left.remapped,cam_right.remapped,cam_left.timestamp);
            border_draw(cam_left, cam_right,cam_left.timestamp);
            //enemy_mapping(cam_left.remapped,cam_right.remapped,cam_left.timestamp);
            //hard_coded_cylinders(cam_left.remapped,temp);
            break;
        case 23:
            draw_cylinder_config(cam_right.remapped);
            //enemy_mapping(cam_left.remapped,cam_right.remapped,cam_left.timestamp);
            break;
        case 24:
            //corresponding_objects(cam_left,cam_right,objects,output);
            //update_elements(cam_left.timestamp,objects);
            //objects.clear();
            break;
        default:
            break;
        }
        wrapper::show_image("left",cam_left.output);
	wrapper::show_image("left_rectified",cam_left.remapped);
        wrapper::show_image("right",cam_right.output);
	wrapper::show_image("right_rectified",cam_right.remapped);
       //wrapper::show_image("edges",cam_left.edges);
        pthread_barrier_wait(&barrier_correspondence);
        wrapper::show_image("disparity",output);
        //cvShowImage("error",raw_output2);
	//key = 0;
        key = cvWaitKey(20);

        if (key == 'q')
        {
            send_stop_to_self();
            stop=1;
        }
        else if (key == 's')
        {
            saveImage(cam_left.remapped.raw(),cam_right.remapped.raw());
            cvSaveImage("disparity_output.png",raw_output);
            //cvSaveImage("location.png",location);
        }
        else if (key == 'c')
        {
            printf("storing color recognition matrix\n");
            cvSaveImage("exclusion.png",exclusion);
        }
        else if(key == 'y') {
            alter_focal_length(+0.01);
        }
        else if(key == 'h') {
            alter_focal_length(-0.01);
        }
        else if(key == 'u') {
            alter_view_height(+0.01);
        }
        else if(key == 'j') {
            alter_view_height(-0.01);
        }
        else if(key == 'i') {
            alter_angle(+0.5);
        }
        else if(key == 'k') {
            alter_angle(-0.5);
        }
        counter++;
        if (options.input)
        {
            stop=1;
            cvWaitKey(0);
        }
    }
    thread_cleanup();
    printf("freeing memory\n");
    release_matrices(cam_left);
    release_matrices(cam_right);
    printf("all done.\n");
    cvDestroyWindow("disparity");
    cvDestroyWindow("left");
    cvDestroyWindow("right");
    cvDestroyWindow("edges");
    pthread_exit(NULL);
}
