#include "position.hpp"
#include <algorithm>
#include <iostream>
#include <cmath>

using namespace std;
pthread_mutex_t our_position_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t enemy_position_mutex = PTHREAD_MUTEX_INITIALIZER;
std::deque<robot_position> our_position_list;
std::deque<robot_position> enemy_position_list;
bool timeval_compare(const timeval& a, const timeval& b) {
    if(a.tv_sec < b.tv_sec) {
        return a.tv_usec < b.tv_usec;
    } else {
        return false;
    }
}

unsigned long getTimestamp(timeval &input) {
    return (input.tv_sec*1000000+input.tv_usec);
}


void camera_positions(robot_position &rotation_axis, robot_position &left_camera_pos, robot_position &right_camera_pos) {
    // the cameras are mounted directly infront of the rotational axis
    // cameras are 130 mm infront of axis, but center-point is approx 20 mm further away from axis.
    const int distance_from_rotation_axis = 175; // 12 cm to camera back, 5.5 cm longer to front of cameras. (rotation is 8 cm from test setup)
    const int left_camera = -40; // testing - axis is about 3 cm to the left (no idea why)
    const int right_camera = +40; // testing
    double middle_y = sin(rotation_axis.psi)*distance_from_rotation_axis;
    double middle_x = cos(rotation_axis.psi)*distance_from_rotation_axis;
    double rel_y = sin(rotation_axis.psi+(90*M_PI/180.0));
    double rel_x = cos(rotation_axis.psi+(90*M_PI/180.0));

    // setting timestamp and psi to same as rotation axis.
    left_camera_pos=rotation_axis;
    right_camera_pos=rotation_axis;

    // shifting camera positions
    left_camera_pos.y=rotation_axis.y+middle_y+rel_y*left_camera;
    left_camera_pos.x=rotation_axis.x+middle_x+rel_x*left_camera;
    right_camera_pos.y=rotation_axis.y+middle_y+rel_y*right_camera;
    right_camera_pos.x=rotation_axis.x+middle_x+rel_x*right_camera;
}


void OurPositionAdd(robot_position& element) {
    pthread_mutex_lock(&our_position_mutex);
    our_position_list.push_front(element);
    // we get 100 readings per sec, and each match is 90 seconds, so 9000 is a nice upper bound.
    // this is only useful while testing, as running this for hours could cause the queue to get extremely large.
    // it is highly unlikely that data for more than a few seconds will be useful.
    // 1000 is 10 seconds of data at 100 msg/sec
    while(our_position_list.size() >= 1000) {
        our_position_list.pop_back();
    }
    pthread_mutex_unlock(&our_position_mutex);
}

void EnemyPositionAdd(robot_position& element) {
    pthread_mutex_lock(&enemy_position_mutex);
    enemy_position_list.push_front(element);
    // we get 100 readings per sec, and each match is 90 seconds, so 9000 is a nice upper bound.
    // this is only useful while testing, as running this for hours could cause the queue to get extremely large.
    // it is highly unlikely that data for more than a few seconds will be useful.
    // 1000 is 10 seconds of data at 100 msg/sec
    while(enemy_position_list.size() >= 1000) {
        enemy_position_list.pop_back();
    }
    pthread_mutex_unlock(&enemy_position_mutex);
}

robot_position getOurPosInterpolated(unsigned long& desired_timestamp) {
    timeval tp;
    tp.tv_sec = desired_timestamp/1000000;
    tp.tv_usec = desired_timestamp%1000000;
    return getOurPosInterpolated(tp);
}

robot_position getEnemyPosInterpolated(unsigned long& desired_timestamp) {
    timeval tp;
    tp.tv_sec = desired_timestamp/1000000;
    tp.tv_usec = desired_timestamp%1000000;
    return getEnemyPosInterpolated(tp);
}


robot_position getOurPosInterpolated(timeval& desired_timestamp) {
    // this is a simple linear interpolation between two values.
    // The speed of the robot is far from constant, so it's definitely not accurate,
    // but better than no interpolation at all.
    // improvements could be done by using a larger sample of points to derivate acceleration etc.
    robot_position retval;
    // waiting to sync data.
    pthread_mutex_lock(&our_position_mutex);
    retval = getPosInterpolated(desired_timestamp,our_position_list);
    pthread_mutex_unlock(&our_position_mutex);
    return retval;
}

robot_position getEnemyPosInterpolated(timeval& desired_timestamp) {
        // this is a simple linear interpolation between two values.
    // The speed of the robot is far from constant, so it's definitely not accurate,
    // but better than no interpolation at all.
    // improvements could be done by using a larger sample of points to derivate acceleration etc.
    robot_position retval;
    // waiting to sync data.
    pthread_mutex_lock(&enemy_position_mutex);
    retval = getPosInterpolated(desired_timestamp,enemy_position_list);
    pthread_mutex_unlock(&enemy_position_mutex);
    return retval;
}

// this is *NOT* threadsafe, use through helper functions
robot_position getPosInterpolated(timeval& desired_timestamp,deque<robot_position> &positions_list) {
    robot_position retval;
    if(getTimestamp(desired_timestamp) == 0.0) {
        cout << "desired timestamp was 0, feeding manual position" << endl;
        retval.timestamp=desired_timestamp;
        retval.y=30;
        retval.x=2400;
        retval.psi=90.0*M_PI/180.0;
        return retval;
    }
    retval.timestamp = desired_timestamp;
    robot_position last_pos,first_pos;
    cout << "desired timestamp: " << getTimestamp(desired_timestamp) << endl;
    for(int i=1; i+1<positions_list.size(); i++) {
        if(getTimestamp(positions_list[i].timestamp)<getTimestamp(desired_timestamp)) {
            first_pos = positions_list[i];
            //printf("ID: %d size: %d\n", i-1, (int) positions_list.size());
            last_pos = positions_list.at(i-1);
            cout << i << " diff(first): " << getTimestamp(desired_timestamp)-getTimestamp(first_pos.timestamp) << endl;
            if(getTimestamp(desired_timestamp)>getTimestamp(last_pos.timestamp)) {
                cout << "ERROR: Current reading from Navsys is older than captured frame!" << endl;
                return last_pos;
            } else {
                cout << i << " diff(last): " << getTimestamp(last_pos.timestamp)-getTimestamp(desired_timestamp) << endl;
            }
            break;
        } else if(i+2 == positions_list.size()) {
            cout << "BORK!" << endl;
            return retval;
        }
    }
    // first we determine how close we are to the data points.
    unsigned long desired_time = getTimestamp(desired_timestamp);
    unsigned long first_pos_time = getTimestamp(first_pos.timestamp);
    unsigned long last_pos_time = getTimestamp(last_pos.timestamp);
    double percentage=0.0;
    if(last_pos_time-first_pos_time != 0) {
    	percentage = (desired_time-first_pos_time)/(last_pos_time-first_pos_time);
    }
    retval.y = (last_pos.y-first_pos.y)*percentage+first_pos.y;
    retval.x = (last_pos.x-first_pos.y)*percentage+first_pos.x;
    // orientation is tricky, requires testing.
    retval.psi = (last_pos.psi-first_pos.psi)*percentage+first_pos.psi;
    cout << "Position: Y: " << retval.y << " X: " << retval.x << " Angle: " << retval.psi*180.0/M_PI << endl;
    return retval;
}

