#include "game_elements.hpp"
#include "../wrapper.hpp"
#include "../segmentation.hpp"
#include "../correspondence.hpp"
#include "../ground_truth.hpp"
#include <sstream>
/* This routine is specific for finding game elements (corns and tomatoes)

    There are essentially two viable approaches:
    - use edges and using symmetry find the middle of the game_elements and use Hough Lines to find position.
    - use segmentation, trace around edges, finding centre, shape etc.
*/
using namespace std;


IplConvKernel* element = 0;
int element_shape = CV_SHAPE_ELLIPSE;
ground_truth img_gt;

void initialize_element()
{
    element = cvCreateStructuringElementEx( 4*2+1, 4*2+1, 4, 4, CV_SHAPE_ELLIPSE, 0 ); // CV_SHAPE_ELLIPSE
}

void find_region_color(RgbImage& image, property& input)
{
    CvRect bbox = input.bounding_box;
    int x=0;
    RgbPixelInt accum; // this will overflow if the region has over 8.8 million points, and they are all white. (impossible with 1024x768 atleast)
    accum.r=0;
    accum.g=0;
    accum.b=0;
    long counter=0;
    for (int y=bbox.y; y<bbox.height+bbox.y; y++)
    {
        if (input.horizontal_map.count(y)>0)
        {
            x=input.horizontal_map[y].at(0);
            accum.r+=image[y][x+1].r;
            accum.g+=image[y][x+1].g;
            accum.b+=image[y][x+1].b;
            counter++;
        }
    }
    input.color.r = round(accum.r/(double) counter);
    input.color.g = round(accum.g/(double) counter);
    input.color.b = round(accum.b/(double) counter);
}


// this routine creates a map of x positions for every y position in a contour.
// it also calculates the center of the region,
// and stores the minimum rotated rectangle coordinates.
// Simple shapes such as balls and cylinders,
// will typically only have two x positions (left and right) for every vertical line.
// this is used for disparity calculations
void create_map(property& input)
{
    input.areal = cvContourArea( &input.contour, CV_WHOLE_SEQ);

    // minimum rotated rectangle (uses rotating calipers)
    CvBox2D box = cvMinAreaRect2( &input.contour, 0 );
    cvBoxPoints( box, input.min_box);
    input.center.x = cvRound((input.min_box[0].x+input.min_box[2].x)/2.0);
    input.center.y = cvRound((input.min_box[0].y+input.min_box[2].y)/2.0);
    CvPoint p;
    for (int i=0; i<input.contour.total; i++)
    {
        p = *CV_GET_SEQ_ELEM(CvPoint,&input.contour,i);
        if (input.horizontal_map.count(p.y)==0)
        {
            vector<unsigned short> element;
            element.push_back(p.x);
            input.horizontal_map.insert(pair<const unsigned short,vector< unsigned short > >(p.y,element));
        }
        else
        {
            input.horizontal_map[p.y].push_back(p.x);
        }
    }
}

CvScalar to_scalar(RgbPixel& input)
{
    return cvScalar(input.b,input.g,input.r);
}

bool is_within_board(vector<short> &map_out, CvRect bbox)
{
    // this routine checks if the contour is on the board.
    // one simple way to do this, is to check if the bottom of the bounding box is within the borders of the playing field.

    // remember that the top of the image is 0, while the bottom is 768.
    if (bbox.y+bbox.height > map_out.at(bbox.x) && bbox.y+bbox.height > map_out.at(bbox.x+bbox.width))
    {
        return true;
    }
    else
    {
        return false;
    }
}

// this is the routine that finds which objects correspond with eachother in the left and right image.
// the output should be a list of tomatoes, oranges and cylinders,
// as well as an output image, where the matched contours in the left and
// right image are superimposed on eachother, and distances are printed.
int corresponding_objects(frame<RgbPixel> &left, frame<RgbPixel> &right, vector< CvPoint2D32f > &tomatoes, vector< CvPoint2D32f > &oranges, vector< CvPoint2D32f > &corns, RgbImage& output)
{
    int max_disparity=135; // k/135 = approx 40 cm. (k=4800-5200)
    int min_disparity=17; // k/17 = approx 3 m.

    // build map
    vector<short> left_map_out, right_map_out;
    left_map_out.assign(left.ch1.getWidth(),left.ch1.getHeight()); // assign lowest figure to all vertical lines.
    for (int y=0; y<left.ch1.getHeight(); y++)
    {
        for (int x=0; x<left.ch1.getWidth(); x++)
        {
            if (left.ch1[y][x] != 0)
            {
                left_map_out.at(x)=y;
                //left.remapped[y][x].b=255;
            }
        }
    }
    right_map_out.assign(right.ch1.getWidth(),1024);
    for (int y=0; y<right.ch1.getHeight(); y++)
    {
        for (int x=0; x<right.ch1.getWidth(); x++)
        {
            if (right.ch1[y][x] != 0)
            {
                right_map_out.at(x)=y;
                //right.remapped[y][x].b=255;
            }
        }
    }
    // color profiles
    CvFont font1;
    CvFont font2;
    cvInitFont(&font1,CV_FONT_VECTOR0,0.5,0.5,0,1);
    cvInitFont(&font2,CV_FONT_VECTOR0,0.35,0.35,0,1);
    int counter=0;
    CvPoint pt,pt0;
    for (int i=0; i<left.contours.size(); i++)
    {
        // first doing some image location exclusion
        CvRect left_bbox = left.contours.at(i).bounding_box;
        if (left_bbox.x < 25)
        {
            // object is too close to border
            continue;
        }
        // this code is for mapping out stuff not on the table, requires input from navigation system.
        //if(!is_within_board(left_map_out,left_bbox)) {
        //     continue;
        //}
        RgbPixel left_color = left.contours.at(i).color;
        printf("left contour: %d, R: %d G: %d B: %d\n", i, left_color.r,left_color.g,left_color.b);
        int matches=0;
        vector<int> possible_matches;
        int match_index=0;

        for (int j=0; j<right.contours.size(); j++)
        {
            CvRect right_bbox = right.contours.at(j).bounding_box;

            // this code is for mapping out stuff not on the table, requires input from navigation system.
            //if(!is_within_board(right_map_out,right_bbox)) {
            //    continue;
            //}
            if (left.contours.at(i).center.x<right.contours.at(j).center.x+max_disparity && right.contours.at(j).center.x+min_disparity < left.contours.at(i).center.x)
            {
                if (right.contours.at(j).center.y < left.contours.at(i).center.y+min_disparity && right.contours.at(j).center.y > left.contours.at(i).center.y-min_disparity)
                {
                    double diff = RGB_color_diff(left_color,right.contours.at(j).color);
                    //printf("%f\n",diff);
                    if (right_bbox.x+right_bbox.width >= 1000)
                    {
                        // object is too close to border to be trusted, continuing.
                        continue;
                    }
                    if (diff < 20.0)
                    {
                        possible_matches.push_back(j);
                    }
                }
            }
        }

        if (possible_matches.size()>0)
        {
            /*
            cout << "contour: " << i << " matches contours:" << flush;
            for (int p=0; p<possible_matches.size(); p++)
            {
                cout << possible_matches.at(p) << " " << flush;
            }
            cout << endl;
            */
            RgbPixel right_color = right.contours.at(possible_matches.at(0)).color;
            CvRect right_bbox = right.contours.at(possible_matches.at(0)).bounding_box;
            vector<int> left_disparity;
            vector<int> right_disparity;
            left_disparity.assign(200,0);
            right_disparity.assign(200,0);
            //printf("WTF: %d\n", left_bbox.y);
            for (int l_y=left_bbox.y; l_y<left_bbox.height+left_bbox.y; l_y++)
            {
                //cout << "GOING_THROUGH" << endl;
                if (left.contours.at(i).horizontal_map.count(l_y)>0)
                {
                    if (right.contours.at(possible_matches.at(0)).horizontal_map.count(l_y)>0)
                    {
                        int disp = left.contours.at(i).horizontal_map[l_y].at(0)-right.contours.at(possible_matches.at(0)).horizontal_map[l_y].at(0);
                        //cout << "DISPARITY: " << disp << endl;
                        if (disp > min_disparity && disp < max_disparity)
                        {
                            left_disparity.at(disp)++;
                        }
                        int l_size = left.contours.at(i).horizontal_map[l_y].size();
                        int r_size = right.contours.at(possible_matches.at(0)).horizontal_map[l_y].size();
                        disp = left.contours.at(i).horizontal_map[l_y].at(l_size-1)-right.contours.at(possible_matches.at(0)).horizontal_map[l_y].at(r_size-1);
                        if (disp > min_disparity && disp < max_disparity)
                        {
                            right_disparity.at(disp)++;
                        }
                    }
                }
            }
            int max_hit=0;
            int left_hit_idx=0;
            for (int d=0; d<left_disparity.size(); d++)
            {
                if (left_disparity[d] > max_hit)
                {
                    max_hit=left_disparity[d];
                    left_hit_idx=d;
                }
            }
            max_hit=0;

            int right_hit_idx=0;
            for (int d=0; d<right_disparity.size(); d++)
            {
                if (right_disparity[d] > max_hit)
                {
                    max_hit=right_disparity[d];
                    right_hit_idx=d;
                }
            }
            int hit_idx=0;
            if (right_hit_idx > left_hit_idx)
            {
                hit_idx = left_hit_idx;
            }
            else
            {
                hit_idx = right_hit_idx;
            }
            double distance = disparity_to_distance(hit_idx);
            //printf("Disparity: %d, Distance: %f\n",hit_idx, disparity_to_distance(hit_idx));
            CvPoint2D32f left_center;
            if (hit_idx > min_disparity && hit_idx < max_disparity)
            {
                if (left.contours.at(i).contour.total>hit_idx && right.contours.at(possible_matches.at(0)).contour.total>hit_idx)
                {
                    // output least fitting rectangle
                    /*
                    pt0.x = cvRound(left.contours.at(i).min_box[3].x);
                    pt0.y = cvRound(left.contours.at(i).min_box[3].y);
                    for ( int var = 0; var < 4; var++ )
                    {
                        pt.x = cvRound(left.contours.at(i).min_box[var].x);
                        pt.y = cvRound(left.contours.at(i).min_box[var].y);
                        cvLine(left.remapped.raw(), pt0, pt, CV_RGB(0, 0, 255), 1, CV_AA, 0);
                        pt0 = pt;
                    }
                    */
                    /*
                    pt.x = cvRound((left.contours.at(i).min_box[0].x+left.contours.at(i).min_box[2].x)/2.0);
                    pt.y = cvRound((left.contours.at(i).min_box[0].y+left.contours.at(i).min_box[2].y)/2.0);
                    pt0.x = cvRound((left.contours.at(i).min_box[1].x+left.contours.at(i).min_box[3].x)/2.0);
                    pt0.y = cvRound((left.contours.at(i).min_box[1].y+left.contours.at(i).min_box[3].y)/2.0);
                    cvLine(left.remapped.raw(), pt0, pt, CV_RGB(255,0,0), 2,CV_AA,0);
                    */
                    //double pos_x = img_gt.horizontal_position_from_distance(distance,cvRound(left.contours.at(i).center.x));
                    double left_x_value = left.contours.at(i).center.x;
                    double right_x_value = right.contours.at(possible_matches.at(0)).center.x;
                    double left_pos_x = (left_x_value-(output.getWidth()/2.0))*2.0*distance_to_view_x(distance)/output.getWidth();
                    double right_pos_x = (right_x_value-(output.getWidth()/2.0))*2.0*distance_to_view_x(distance)/output.getWidth();
                    double pos_x = (left_pos_x+right_pos_x)/2.0;
                    double pos_difference = (left_pos_x-right_pos_x-8.50); // 8.50 is the difference between cameras.


                    // y combined with distance is used to determine what is on the table, and what is not on the table.
                    double left_y_value = left.contours.at(i).center.y;
                    double right_y_value = right.contours.at(possible_matches.at(0)).center.x;
                    double left_pos_y = (left_y_value-(output.getHeight()/2.0))*2.0*distance_to_view_y(distance)/output.getHeight();
                    double right_pos_y = (right_y_value-(output.getHeight()/2.0))*2.0*distance_to_view_y(distance)/output.getHeight();



                    //double distance_at_ground = sqrt(distance*distance-(35-5)*(35-5)); // using pythagoras, a^2+b^2=c^2, a=sqrt(c^2-b^2) (distance-height)
                    double distance_at_ground = distance+8.0;
                    double pos_x_fixed = pos_x-4.0;
                    CvPoint p;
                    stringstream distance_data;
                    distance_data << distance_at_ground;
                    //distance_data << distance_at_ground;
                    stringstream position_data;
                    position_data << " (";
                    position_data << pos_x;
                    position_data << " )";
                    string distance_text = distance_data.str();
                    string position_text = position_data.str();
                    //cout << "Distance: " << distance << " Left(X): " << left_pos_x << " Right(X): " << right_pos_x << " Difference: " << pos_difference << endl;
                    // calculate position to put text
                    p.x = cvRound((left.contours.at(i).center.x+right.contours.at(possible_matches.at(0)).center.x)/2.0)-10;
                    p.y = cvRound((left.contours.at(i).center.y+right.contours.at(possible_matches.at(0)).center.y)/2.0);

                    // Add object to list
                    // object classification based on color
                    // red = tomatoes, orange = oranges, white = corns

                    CvPoint2D32f object;
                    object.y = distance_at_ground;
                    object.x = pos_x_fixed;
                    double rg_ratio = (left_color.r+1.0)/(left_color.g+1.0);
                    double gb_ratio = (left_color.g+1.0)/(left_color.b+1.0);
                    CvScalar draw_color;
                    if (rg_ratio > 1.8)
                    {
                        // orange or tomato
                        if (gb_ratio < 1.5)
                        {
                            tomatoes.push_back(object);
                        }
                        else
                        {
                            oranges.push_back(object);
                        }
                    }
                    else
                    {
                        corns.push_back(object);
                    }

                    // drawing contours
                    cvDrawContours(output.raw(),&left.contours.at(i).contour,to_scalar(left_color),to_scalar(left_color),0,CV_FILLED,8,cvPoint(0,0));
                    cvDrawContours(output.raw(),&right.contours.at(possible_matches.at(0)).contour,to_scalar(right_color),to_scalar(right_color),0,CV_FILLED,8,cvPoint(0,0));
                    cvPutText( output.raw(), distance_text.c_str(), p, &font1, CV_RGB(255,255,255));
                    p.y+=20;
                    cvPutText( output.raw(), position_text.c_str(), p, &font2, CV_RGB(128,128,255));

                }
            }
            //printf("LEFT: %d %d, RIGHT: %d %d\n",left_bbox.y,left_bbox.x,right_bbox.y,right_bbox.x);
        }
    }
    left.contours.clear();
    right.contours.clear();

    // WARNING: this memory structure is used by CvSeq, which means that even though I feel it should be possible to release it earlier,
    // it fails miserably.
    cvReleaseMemStorage(&left.storage);
    cvReleaseMemStorage(&right.storage);
}

// this routine calls the various methods to find the region contours and their properties,
// as well as doing the color thresholding and calling the color segmentation
int find_game_elements_individual(frame<RgbPixel> &camera)
{
    initialize_element();
    //cvSmooth(camera.remapped.raw(),camera.remapped.raw(),CV_MEDIAN,3,3);
    for(int y=0; y<camera.remapped.getHeight(); y++) {
       for(int x=0; x<camera.remapped.getWidth(); x++) {
                normalize_fix(camera.diff[y][x],0.0);
      }
    }

    // this removes all pixels where red is not the dominant pixel value.
    // Only oranges and red balls should be matched on the playing field (not green background or cylinders).
    // It is dependant on a proper white balance, and with automatic white balance it does not work properly.
    // Sane white balance values must be set in camera.cpp
    for (int y=0; y<camera.remapped.getHeight(); y++)
    {
        for (int x=0; x<camera.remapped.getWidth(); x++)
        {
            double rg_ratio = (camera.remapped[y][x].r+1)/(double) (camera.remapped[y][x].g+1);
            double bg_ratio = (camera.remapped[y][x].b+1)/(double) (camera.remapped[y][x].g+1);
            if (rg_ratio <= 1.0)
            {
                camera.remapped[y][x].r=0;
                camera.remapped[y][x].g=0;
                camera.remapped[y][x].b=0;
            }
        }
    }
    wrapper::copy_image(camera.remapped,camera.diff);

    // these two routines combined perform a Morphological "Opening"
    cvErode(camera.remapped.raw(),camera.remapped.raw(),element,1);
    cvDilate(camera.remapped.raw(),camera.remapped.raw(),element,1);

    // this is designed to only match white cylinders on the playing field.
    // the result is drawn into the same image as the red balls.
    for (int y=0; y<camera.remapped.getHeight(); y++)
    {
        for (int x=0; x<camera.remapped.getWidth(); x++)
        {
            double rg_ratio = (camera.remapped[y][x].r+1)/(double) (camera.remapped[y][x].g+1);
            if (rg_ratio >= 1.5) // trying to avoid seeing yellow
            {
                // this matches red (ish)
            }
            else if (intensity(camera.diff[y][x])>120.0 && rg_ratio >= 0.99)
            {
                camera.remapped[y][x]=camera.diff[y][x];
                // this matches white
            }
            else
            {
                camera.remapped[y][x].r=0;
                camera.remapped[y][x].g=0;
                camera.remapped[y][x].b=0;
            }
        }
    }
    // this performs color segmentation, first by doing a RGB gradient, followed by the scanline color segmentation algorithm.
    RGB_Gradient(camera.remapped.raw(), camera.dx,2);
    MatShort gradient(camera.dx);
    color_encode(camera.remapped, camera.dx, camera.output,70, 70); // was: 70, 70;

    // Threshold the source image. This needful for cvFindContours().
    wrapper::convert_rgb2gray(camera.output,camera.ch1);
    cvThreshold( camera.ch1.raw(), camera.ch1.raw(), 0, 255, CV_THRESH_BINARY );

    // Find all contours.
    camera.storage = cvCreateMemStorage(0);
    CvSeq* contours = cvCreateSeq(CV_SEQ_ELTYPE_POINT, sizeof(CvSeq), sizeof(CvPoint) , (CvMemStorage*) camera.storage);

    // this OpenCV routine essentially performs border tracing, but it also removes contours within contours,
    // and puts it in a data format that is easy to deal with.
    cvFindContours( camera.ch1.raw(), camera.storage, &contours, sizeof(CvContour),
                    CV_RETR_EXTERNAL, CV_CHAIN_APPROX_NONE, cvPoint(0,0)); // CV_RETR_LIST for all contours, CV_RETR_EXTERNAL for external.

    // calculate some individual properties of contours
    for (;contours;contours = contours->h_next)
    {
        if (contours->total < 20)
        {
            continue;
        }
        property temp;
        temp.contour = *contours;
        temp.bounding_box = cvBoundingRect( contours, 0 );
        create_map(temp);
        find_region_color(camera.remapped,temp);
        camera.contours.push_back(temp);
    }
}
