#include "segmentation.hpp"
#include "correspondence.hpp"
#include "conversion.hpp"
#include <cstdio>
using namespace std;


// This is a data structure used for checking borders of a pixel,
// and it makes checking neighbouring pixels much easier.
coord direction[8];
void init_direction()
{
    // 0 1 2
    // 7   3
    // 6 5 4
    direction[0].x = -1;
    direction[0].y = -1;
    direction[1].x = 0;
    direction[1].y = -1;
    direction[2].x = 1;
    direction[2].y = -1;
    direction[3].x = 1;
    direction[3].y = 0;
    direction[4].x = 1;
    direction[4].y = 1;
    direction[5].x = 0;
    direction[5].y = 1;
    direction[6].x = -1;
    direction[6].y = 1;
    direction[7].x = -1;
    direction[7].y = 0;
}

double intensity(RgbPixel &pixel)
{
    return sqrt((double) (pixel.r*pixel.r) + (pixel.g*pixel.g) + pixel.b*pixel.b);
}

typedef struct
{
    short I1, I2, I3;
} seg_features;

short calc_I1(RgbPixel& pixel)
{
    return (pixel.r+pixel.g+pixel.b)/3;
}

short calc_I2(RgbPixel& pixel)
{
    return (pixel.r-pixel.b)/2;
}

short calc_I3(RgbPixel& pixel)
{
    return (2*pixel.g-pixel.r-pixel.b)/4;
}

short calc_grayness(RgbPixel& pixel)
{
    return (abs(pixel.r-pixel.g) + abs(pixel.r-pixel.b) + abs(pixel.g-pixel.b))/2; // 0-512, scaled to 0-255.
}

double calc_grayness_ratio(RgbPixel& pixel)
{
    double ratio_rg = (pixel.r+1)/(pixel.g+1);
    double ratio_bg = (pixel.b+1)/(pixel.g+1);
    return max(ratio_rg, ratio_bg);
}

RgbPixel allSet(unsigned char value)
{
    RgbPixel p;
    p.r=value;
    p.g=value;
    p.b=value;
    return p;
}

void setval(RgbPixel& p, unsigned char r, unsigned char g, unsigned char b)
{
    p.r=r;
    p.g=g;
    p.b=b;
}


// This is used to separate Hue values into colors. It expects Hue/30 as input.
// Values 20 and 21 are used for white and black.
RgbPixel get_color(unsigned char value)
{
    RgbPixel p;
    switch (value)
    {
    case 0:
        setval(p,255,0,0); // red
        break;
    case 1:
        setval(p,255,255,0); // yellow
        break;
    case 2:
        setval(p,0,128,0); // green
        break;
    case 3:
        setval(p,0,192,0); // green
        break;
    case 4:
        setval(p,0,255,0); // green
        break;
    case 5:
        setval(p,0,255,255); // cyan
        break;
    case 6:
        setval(p,0,0,255); // blue
        break;
    case 7:
        setval(p,0,0,255); // blue
        break;
    case 8:
        setval(p,0,0,255); // blue
        break;
    case 9:
        setval(p,255,0,255); // magenta
        break;
    case 10:
        setval(p,255,0,0); // red
        break;
    case 11:
        setval(p,255,0,0); // red
        break;
    case 20:
        setval(p,255,255,255); // white
        break;
    case 21:
        setval(p,0,0,0); // black
        break;
    default:
        setval(p,128,128,128); // unknown = grey
        break;

    }
    return p;
}

unsigned char pixel_detect(RgbPixel& input)
{
    short hue=HuefromRGB(input);
    unsigned char color_val=hue/30;
    if (calc_grayness(input)<70)
    {
        return 255;
    }
    else
    {
        return color_val;
    }

    // red: I1: 10  125
    // green: I1: 60    160 (30 in shade)
    //
    // white: I1: 100   256 (saturation)
    // black: I1: 0     60
    /*
    if(f.I2 > 25 && f.I2 < 128) {
        output[y][x].g=255;
    } else {
        output[y][x].r=f.I1;
        output[y][x].g=f.I2+127;
        output[y][x].b=f.I3+127;
    }
    */

}

typedef struct
{
    short x1, y1, x2, y2;
} box;

void run_length_encode(RgbImage& input, vector< vector <run_len> > &rle) {
    for(int y=0; y<input.getHeight(); y++) {
        for(int x=0; x<input.getWidth(); x++) {
            // going through region.
            while(input[y][x].r<3 && input[y][x].g<3 && input[y][x].b<3 && x<input.getWidth()) {
                x++;
            }
            run_len temp;
            temp.pos=x;
            while(!pixel_is_zero(input[y][x]) && x<input.getWidth()) {
                x++;
            }
            temp.sz = x-temp.pos;
            temp.colorval = input[y][x-1];
            rle[y].push_back(temp);
        }
        /*
        printf("line: %d regions: %d [",y,rle.at(y).size());
        for(int i=0; i<rle.at(y).size(); i++) {
            printf(" %d(%d)",rle.at(y).at(i).pos,rle.at(y).at(i).sz);
        }
        printf(" ]\n");
        */
    }
}

void run_length_encode_diff(RgbImage& input, vector< vector<run_len> > &rle) {
    for(int y=0; y<input.getHeight(); y++) {
        for(int x=0; x<input.getWidth(); x++) {
            run_len temp;
            temp.pos=x;
            while(color_difference(input[y][temp.pos],input[y][x])<10 && x<input.getWidth()) {
                x++;
            }
            temp.sz = x-temp.pos;
            temp.colorval = input[y][temp.pos];
            rle[y].push_back(temp);
        }
    }
}


// Scanline color segmentation
// input is the image to segment, while output is the segmented image.
// The variance should preferably be a RGB Gradient image, or a I1I2I3 gradient image.
// var1 is the threshold in vertical direction,
// while var2 is threshold in horizontal direction.
// good results are achieved with 70 for RGB gradient, and 30-40 for I1I2I3 gradient.
// With cameras with more vibrant colors, higher thresholds might yield better results.
// For good results, smooth_regions should be applied to the output image.

// Too high threshold will cause holes in regions, and possibly over segmentation,
// While too low threshold might cause streaking effects, and ruined segmentation.
// POSSIBLE IMPROVEMENTS:
// - using three gradient images instead of one, one for I1, one for I2, one for I3, and a threshold for each.
// - finding a way to detect streaking line errors.
// - possibly using color difference between pixels further apart
// - testing with gradient strength in only horizontal or vertical direction (instead of a combination, regardless of direction).
void color_encode(RgbImage& input, CvMat* raw_variance, RgbImage& output, int var1, int var2)
{
    MatShort variance(raw_variance);
    RgbPixelInt values;
    //output.zero_fill();
    int x=0,y=0;

    // Scan line, left to right, top to bottom.
    for (x=0; x<input.getWidth(); x++)
    {
        //cout << "line: " << y << endl;
        for (y=0; y<input.getHeight(); y++)
        {
            //if(variance[y][x]<=max_variance) {
            //cout << "match: " << y << " " << x << endl;
            values.r=0;
            values.g=0;
            values.b=0;
            for (int a=0; a+y<input.getHeight(); a++)
            {
                if (variance[y+a][x]>var1 && variance[y+a][x]>var1 || a+y==input.getHeight()-1)
                {
                    for (int b=0; b<a; b++)
                    {
                        output[y+b][x].r=values.r/a;
                        output[y+b][x].g=values.g/a;
                        output[y+b][x].b=values.b/a;
                    }
                    if (a==0)
                    {
                        output[y][x].r=0;
                        output[y][x].g=0;
                        output[y][x].b=0;
                    }
                    y+=a;
                    break;
                }
                else
                {
                    //cout << "." << flush;
                    values.r+=input[y+a][x].r;
                    values.g+=input[y+a][x].g;
                    values.b+=input[y+a][x].b;
                }
            }
            //cout << endl;
            //}
        }
    }

    // scan line, top to bottom, left to right.
    for (y=0; y<input.getHeight(); y++)
    {
        //cout << "line: " << y << endl;
        for (x=0; x<input.getWidth(); x++)
        {
            //if(variance[y][x]<=max_variance) {
            //cout << "match: " << y << " " << x << endl;
            values.r=0;
            values.g=0;
            values.b=0;
            for (int a=0; a+x<input.getWidth(); a++)
            {
                if (variance[y][x+a]>var2 || a+x==input.getWidth()-1)
                {
                    for (int b=0; b<a; b++)
                    {
                        output[y][x+b].r=values.r/a;
                        output[y][x+b].g=values.g/a;
                        output[y][x+b].b=values.b/a;
                    }
                    x+=a;
                    break;
                }
                else
                {
                    //cout << "." << flush;
                    values.r+=output[y][x+a].r;
                    values.g+=output[y][x+a].g;
                    values.b+=output[y][x+a].b;
                }
            }
            //cout << endl;
            //}
        }
    }
}

bool pixel_is_zero(RgbPixel& pixel)
{
    if (pixel.r == 0 && pixel.g == 0 && pixel.b == 0)
    {
        return true;
    }
    else
    {
        return false;
    }
}


// This routine smooths segmented regions using scan lines, first left to right, top to bottom,
// then top to bottom, left to right.
// it essentially just averages pixels between each black pixel in each direction.
// It might be beneficial to apply several times, and especially after fixing regions.
void region_fix(RgbImage& input, RgbImage& original, RgbImage& temp, int region_variance)
{
    RgbPixel previous,avg,next;
    int a=0;
    for (int y=0; y<input.getHeight(); y++)
    {
        for (int x=1; x<input.getWidth(); x++)
        {
            if (pixel_is_zero(input[y][x]))
            {
                previous=input[y][x-1];
                for (a=1; a+x<input.getWidth() && pixel_is_zero(input[y][x+a]); a++)
                {
                }
                next=input[y][x+a];
                if (color_difference(previous, next) > region_variance)
                {
                    for (int b=0; b<a; b++)
                    {
                        int next_diff = color_difference(next, original[y][x+b]);
                        int previous_diff = color_difference(previous, original[y][x+b]);
                        if (next_diff <= previous_diff)
                        {
                            temp[y][x+b]=next;
                        }
                        else
                        {
                            temp[y][x+b]=previous;
                        }
                    }
                }
                x+=a;
            }
        }
    }

    for (int x=0; x<input.getWidth(); x++)
    {
        for (int y=1; y<input.getHeight(); y++)
        {
            if (pixel_is_zero(input[y][x]))
            {
                previous=input[y-1][x];
                for (a=1; a+y<input.getHeight() && pixel_is_zero(input[y+a][x]); a++)
                {
                }
                next=input[y+a][x];
                if (color_difference(previous, next) > region_variance)
                {
                    for (int b=0; b<a; b++)
                    {
                        int next_diff = color_difference(next, original[y+b][x]);
                        int previous_diff = color_difference(previous, original[y+b][x]);
                        if (next_diff <= previous_diff)
                        {
                            //if(color_difference(temp[y+b][x], next) < region_variance) {
                            temp[y+b][x]=next;
                            //}
                        }
                        else
                        {
                            //if(color_difference(temp[y+b][x],previous) < region_variance) {
                            temp[y+b][x]=previous;
                            //}
                        }
                    }
                }
                y+=a;
            }
        }
    }
}


// this routine is used to smooth regions that are already segmented with the scanline color segmentation.
// It passes through the image, left to right, top to bottom, and between each black pixel, averages the pixel values.
// It then proceeds to do the same, but top to bottom, left to right.
// This is especially useful if some form of region fixing has been applied.
void smooth_region(RgbImage& input)
{
    RgbPixelInt values;
    int a=0;
    for (int x=0; x<input.getWidth(); x++)
    {
        for (int y=0; y<input.getHeight(); y++)
        {
            if (!pixel_is_zero(input[y][x]))
            {
                values.r=0;
                values.g=0;
                values.b=0;
                for (a=0; a+y<input.getHeight() && !pixel_is_zero(input[y+a][x]); a++)
                {
                    values.r += input[y+a][x].r;
                    values.g += input[y+a][x].g;
                    values.b += input[y+a][x].b;
                }
                for (int b=0; b<a; b++)
                {
                    input[y+b][x].r = values.r/a;
                    input[y+b][x].g = values.g/a;
                    input[y+b][x].b = values.b/a;
                }
                y += a;
            }
        }
    }
    for (int y=0; y<input.getHeight(); y++)
    {
        for (int x=0; x<input.getWidth(); x++)
        {
            if (!pixel_is_zero(input[y][x]))
            {
                values.r=0;
                values.g=0;
                values.b=0;
                for (a=0; a+x<input.getWidth() && !pixel_is_zero(input[y][x+a]); a++)
                {
                    values.r += input[y][x+a].r;
                    values.g += input[y][x+a].g;
                    values.b += input[y][x+a].b;
                }
                for (int b=0; b<a; b++)
                {
                    input[y][x+b].r = values.r/a;
                    input[y][x+b].g = values.g/a;
                    input[y][x+b].b = values.b/a;
                }
                x += a;
            }
        }
    }
}

void remove_narrow(RgbImage& input)
{

}

// This routine is used to fill holes in a segmented image.
// It only fills pixels where the first non-black pixel above,
// below as well as left and right of the pixel is within a certain region variance.
// In practise it fills holes rather well, but it will also in many cases merge regions of similar color.
void fill_holes(RgbImage& input, RgbImage& vertical_temp, int region_variance)
{
    wrapper::zero_fill(vertical_temp.raw());
    RgbPixel previous,next,avg,final_avg;
    int a=0;
    for (int y=0; y<input.getHeight(); y++)
    {
        for (int x=1; x<input.getWidth(); x++)
        {
            if (pixel_is_zero(input[y][x]))
            {
                previous=input[y][x-1];
                for (a=1; a+x<input.getWidth() && pixel_is_zero(input[y][x+a]); a++)
                {
                }
                next=input[y][x+a];
                if (color_difference(previous, next) < region_variance)
                {
                    avg.r=(previous.r+next.r)/2;
                    avg.g=(previous.g+next.g)/2;
                    avg.b=(previous.b+next.b)/2;
                    for (int b=0; b<a; b++)
                    {
                        vertical_temp[y][x+b]=avg;
                    }
                }
                x+=a;
            }
        }
    }
    for (int x=0; x<input.getWidth(); x++)
    {
        for (int y=1; y<input.getHeight(); y++)
        {
            if (pixel_is_zero(input[y][x]))
            {
                previous=input[y-1][x];
                for (a=1; a+y<input.getHeight() && pixel_is_zero(input[y+a][x]); a++)
                {
                }
                next=input[y+a][x];
                if (color_difference(previous, next) < region_variance)
                {
                    avg.r=(previous.r+next.r)/2;
                    avg.g=(previous.g+next.g)/2;
                    avg.b=(previous.b+next.b)/2;
                    for (int b=0; b<a; b++)
                    {
                        if (color_difference(avg,vertical_temp[y+b][x]) < region_variance)
                        {
                            input[y+b][x].r=(avg.r+vertical_temp[y+b][x].r)/2;
                            input[y+b][x].g=(avg.g+vertical_temp[y+b][x].g)/2;
                            input[y+b][x].b=(avg.b+vertical_temp[y+b][x].b)/2;
                        }
                    }
                }
                y+=a;
            }
        }
    }
}

// this routine simply marks the borders of regions with blue.
void border_detect(RgbImage& input, RgbImage& output) {
    init_direction();
	for(int y=1; y<input.getHeight()-1; y++) {
		for(int x=1; x<input.getWidth()-1; x++) {
			if(pixel_is_zero(input[y][x])) {
				for(int a=0; a<8; a++) {
					if(!pixel_is_zero(input[y+direction[a].y][x+direction[a].x])) {
					    //cout << "Mark: " << y << " " << x << endl;
						output[y][x].r=0;
						output[y][x].b=255;
						output[y][x].g=0;
					}
				}
			}
		}
	}
}

// This is a basic implementation of moore border tracing.
// The implementation is rather straight forward,
// even though iterating around neighbouring pixels can make you dizzy at times :)
// see p.332 Sonka&Hlavac
void moore_border_tracing(BwImage& segmentation, BwImage& regions)
{
    // create bounding box
    box bounding_box;
    // Using Moore neighbour tracking to trace edge of contours:

    // first we identify start position.
    int segment=1;
    int y,x;
    int done=0;
    for (y=segmentation.getHeight()-1; y>=0; y--)
    {
        for (x=segmentation.getWidth()-1; x>=0; x--)
        {
            unsigned char p = segmentation[y][x];
            if (p == 0 || p == 10 || p == 11)
            {
                done=1;
                break;
            }
        }
        if (done==1)
        {
            break;
        }
    }
    cout << "DEBUG: " << y << " " << x << endl;
    done=0;
    int startx=x;
    int starty=y;
    int d=3; // iterating over the image bottom->top, right->left, so we are coming from right (3).
    regions[y][x]=segment;
    bounding_box.y1=starty;
    // tracing border of region
    while (!done)
    {
        int match=0;
        for (int a=(d+4+1)%8, i=0; i<8; i++,a = (a+1) % 8)
        {
            short next_x=y+direction[a].y;
            short next_y=x+direction[a].x;
            unsigned char p = segmentation[next_y][next_x];
            if ((p == 0 || p == 10 || p == 11) && regions[next_y][next_x] == 0 )
            {
                regions[next_y][next_x]=segment;
                x += direction[a].x;
                y += direction[a].y;
                //cout << "DEBUG(tracing): " << y << " " << x << endl;
                match=1;
                break;
            }
            if (startx==x+direction[a].x && starty==y+direction[a].y)
            {
                match=0;
                break;
            }
        }
        if (match==0)
        {
            done=1;
        }
    }

}


// This calculates if two pixels are within or outside a color sphere, relative to eachother.
// It is a rather decent metric for color comparison, albeit a bit slow.
bool RGB_color_sphere(RgbPixel& input, RgbPixel& compare, int radius)
{
    unsigned char r_diff, g_diff, b_diff;
    r_diff=abs(input.r-compare.r);
    g_diff=abs(input.g-compare.g);
    b_diff=abs(input.b-compare.b);
    if (sqrt(r_diff*r_diff+g_diff*g_diff+b_diff*b_diff)<= radius)
    {
        return true;
    }
    else
    {
        return false;
    }
}

double RGB_color_diff(RgbPixel& input, RgbPixel& compare) {
    unsigned char r_diff, g_diff, b_diff;
    r_diff=abs(input.r-compare.r);
    g_diff=abs(input.g-compare.g);
    b_diff=abs(input.b-compare.b);
    return sqrt(r_diff*r_diff+g_diff*g_diff+b_diff*b_diff);
}

// This is used for color recognition.
// It tries to detect red, green, white and black parts of the image.
// It doesn't work very well, as the objects, albeit only a single color, vary a lot in intensity.
// Normalization does not work well with the application this was designed for, so this routine is pretty much useless.
void detect_colors(RgbImage& input, RgbImage& output)
{
    BwImage segmentation = cvCreateImage(input.getSize(),8,1);
    BwImage regions = cvCreateImage(input.getSize(),8,1);

    RgbPixel color_red={30,40,131}; // BGR
    RgbPixel color_green={25,220,85};
    RgbPixel color_white={115,190,145}; // BGR
    RgbPixel color_black={1,23,6};
    for (int y=0; y<input.getHeight(); y++)
    {
        for (int x=0; x<input.getWidth(); x++)
        {
            double white=RGB_color_diff(input[y][x],color_white);
            double green=RGB_color_diff(input[y][x],color_green);
            double black=RGB_color_diff(input[y][x],color_black);
            double red=RGB_color_diff(input[y][x],color_red);
            double rg_diff = (input[y][x].r+1)/(double) (input[y][x].g+1);
            segmentation[y][x]=13;
            if(intensity(input[y][x])<80.0) {
                segmentation[y][x]=21;
            } else if(intensity(input[y][x])>200.0) {
                if(rg_diff<0.6) {
                    segmentation[y][x]=3;
                } else {
                    segmentation[y][x]=20;
                }
            } else {
                segmentation[y][x]=13;
                if(input[y][x].g>input[y][x].r) {
                    segmentation[y][x]=3; // 13 = unknown.
                }
                if(input[y][x].g<input[y][x].r) {
                    segmentation[y][x]=0;
                }
            }
                output[y][x]=get_color(segmentation[y][x]);
        }
    }
    /*
    moore_border_tracing(segmentation,regions);
    for(int y=0; y<regions.getHeight(); y++)
    {
        for(int x=0; x<regions.getWidth(); x++)
        {
            if(regions[y][x]==1) {
                setval(output[y][x],255,0,0);
            } else {
                setval(output[y][x],0,0,0);
            }
        }
    }
    */
    int val=0;
    /*
    int old_new;
    for(int y=0; y<input.getHeight(); y++)
    {
        for(int x=0; x<input.getWidth(); x++)
        {
            if(regions[y][x]==0) {
                if(segmentation[y][x]==0 || segmentation[y][x]==10 || segmentation[y][x]==11) {
                    for(int circle_y=0; circle_y+y<input.getHeight(); circle_y++) {
                        for(int circle_x=0; circle_x+x<input.getWidth(); circle_x++) {

                        }
                    }
            }
        }
    }
    */
    // segmentation
}

// This is a naive implementation of Richard E. Blake's Color Segmentation.
// It takes slightly over 1 second to execute on a 1024x768 image,
// so even you could make it 10 times faster, it'd still be rather slow.
void color_segmentation(IplImage* raw_input, IplImage* raw_output, double tolerance, int maxit)
{
    RgbImage input(raw_input);
    RgbImage output(raw_output);
    double var1, var2;
    int it=0;
    int count_changed=0;
    int count_nochange=0;
    bool changed=true;

    while (changed && it<maxit)
    {
        changed=false;
        printf("Iteration: %d (%d max)\n",it,maxit);
        for (int y=0; y< raw_input->height; y++)
        {
            for (int x=0; x< raw_input->width; x++)
            {
                var1=calculate_variance(raw_input,y,x,3);
                var2=calculate_variance(raw_input,y,x,5);
                if (var2<(var1 + tolerance) && var2>0.0)
                {
                    //average
                    output[y][x]=calculate_mean_pixel(raw_input,y,x,3);
                    changed=true;
                    count_changed++;
                }
                else
                {
                    count_nochange++;
                }
            }
        }
        printf("not changed: %d, changed: %d\n",count_nochange, count_changed);
        count_changed=0;
        count_nochange=0;
        it++;
        for (int y=0; y<raw_input->height; y++)
        {
            for (int x=0; x<raw_input->width; x++)
            {
                input[y][x]=output[y][x];
            }
        }
    }
}

// This calculates the mean intensity value of a kernel of size x*y,
// and also works along the image borders.
double calculate_mean(IplImage* raw_input, int y, int x, int masksize)
{
    RgbImage input(raw_input);
    int hm=masksize/2;
    double sum=0;
    int r,g,b;
    int n=masksize*masksize;
    for (int i=y-hm; i<=y+hm; i++)
    {
        for (int j=x-hm; j<=x+hm; j++)
        {
            if (i<0 || j<0 || i>= raw_input->height || j>= raw_input->width)
            {
                n--;
            }
            else
            {
                int r=input[i][j].r*input[i][j].r;
                int g=input[i][j].g*input[i][j].g;
                int b=input[i][j].b*input[i][j].b;
                sum+=sqrt(r+g+b);
            }
        }
    }
    return sum/(double) n;
}

// This calculates the mean pixel value of a kernel of size x*y,
// and also works along the image borders.
RgbPixel calculate_mean_pixel(IplImage* raw_input, int y, int x, int masksize)
{
    RgbImage input(raw_input);
    int hm=masksize/2;
    RgbPixelInt sum;
    RgbPixel retval;
    sum.r=0;
    sum.g=0;
    sum.b=0;
    int n=masksize*masksize;
    for (int i=y-hm; i<=y+hm; i++)
    {
        for (int j=x-hm; j<=x+hm; j++)
        {
            if (i<0 || j<0 || i>= raw_input->height || j>= raw_input->width)
            {
                n--;
            }
            else
            {
                sum.r+=input[i][j].r;
                sum.g+=input[i][j].g;
                sum.b+=input[i][j].b;
            }
        }
    }
    retval.r = sum.r/n;
    retval.g = sum.g/n;
    retval.b = sum.b/n;
    return retval;
}

// This calculates the variance of a kernel of size y * x,
// and also works along image borders.
double calculate_variance(IplImage *raw_input, int y, int x, int masksize)
{
    RgbImage input(raw_input);
    int hm=masksize/2;
    double mean=calculate_mean(raw_input,y,x,masksize);
    double sum=0;
    int r,g,b;
    int n=masksize*masksize;
    for (int i=y-hm; i<=y+hm; i++)
    {
        for (int j=x-hm; j<=x+hm; j++)
        {
            if (i<0 || j<0 || i>= raw_input->height || j>= raw_input->width)
            {
                n--;
            }
            else
            {
                int r=input[i][j].r*input[i][j].r;
                int g=input[i][j].g*input[i][j].g;
                int b=input[i][j].b*input[i][j].b;
                sum+=(sqrt(r+g+b)-mean)*(sqrt(r+g+b)-mean);
            }
        }
    }
    return sum/(double) n;
}

// normalizes color components
void normalize(RgbPixel &pixel, double threshold)
{
    double scale=0.0;
    bool red_biggest=0;
    if (pixel.r>=pixel.g && pixel.r>=pixel.b)
    {
        scale=(255.0/pixel.r);
    }
    if (pixel.g>=pixel.r && pixel.g>=pixel.b)
    {
        scale=(255.0/pixel.g);
    }
    if (pixel.b>=pixel.r && pixel.b>=pixel.g)
    {
        scale=(255.0/pixel.b);
    }
    if (intensity(pixel)>threshold)
    {
        pixel.r *= scale;
        pixel.g *= scale;
        pixel.b *= scale;
    } else {
        pixel.r = 0;
        pixel.g = 0;
        pixel.b = 0;
    }

}


// this does the same as regular normalization,
// except it turns regions below a threshold blue instead of black.
// This is useful for the color segmentation, as it uses black regions as a way to separate regions.
void normalize_fix(RgbPixel &pixel, double threshold)
{
    double scale=0.0;
    bool red_biggest=0, green_biggest=0;
    if (pixel.r>=pixel.g && pixel.r>=pixel.b)
    {
        scale=(255.0/pixel.r);
        red_biggest=1;
    }
    if (pixel.g>=pixel.r && pixel.g>=pixel.b)
    {
        scale=(255.0/pixel.g);
        green_biggest=1;
    }
    if (pixel.b>=pixel.r && pixel.b>=pixel.g)
    {
        scale=(255.0/pixel.b);
    }
    if (intensity(pixel)>threshold)
    {
        pixel.r *= scale;
        pixel.g *= scale;
        pixel.b *= scale;
    } else {
            pixel.r=0;
            pixel.g=0;
            pixel.b=255;
    }

}



// implemented based on the algorithm described in DIP3, p. 813
// must be repeated till the return value is 0.

// Works very nicely, and makes a connected thinned graph of any binary region.
// The performance is extremely variable, and a scene with lots of regions to thin would take a very long time.
// Regular execution is atleast 100 ms before a region is thinned.
int skeletonize(BwImage& input)
{
    init_direction(); // p1 = ?, a%8-1 converts from book to own system.
    int deleted=0;
    int c=0;
    coord val;
    for (int step=1; step<3; step++)
    {
        //cout << "step: " << step << endl;
        vector<coord> flag_for_deletion;
        for (int y=1; y<input.getWidth()-1; y++)
        {
            for (int x=1; x<input.getWidth()-1; x++)
            {
                if (input[y][x] == 255)
                {
                    int n=0; // N(point) (number of neighbours)
                    int t=0; // T(point) (number of transitions)

                    for (int a=0; a<8; a++)
                    {
                        if (input[y+direction[a].y][x+direction[a].x] == 255)
                        {
                            n++;
                        }
                        if (input[y+direction[a].y][x+direction[a].x] < input[y+direction[(a+1)%8].y][x+direction[(a+1)%8].x])
                        {
                            t++;
                        }
                    }
                    // p. 813 DIP3
                    //cout << "skeletonize: " << n << " " << t << endl;
                    if (n >= 2 && n <= 6 // (a)
                            && t == 1 // (b)
                       )
                    {
                        bool rem=false;
                        if (step == 1
                                && (input[y-1][x] * input[y][x+1] * input[y+1][x]) == 0 // (c)
                                && (input[y][x+1] * input[y+1][x] * input[y][x-1]) == 0 // (d)
                                //&& pixels[(w-1) + (h+1)*width].r == 0
                           )
                        {
                            rem=true;
                        }
                        if (step == 2
                                && (input[y-1][x] * input[y][x+1] * input[y][x-1]) == 0 // (c')
                                && (input[y-1][x] * input[y+1][x] * input[y][x-1]) == 0 // (d')
                                //&& pixels[(w+1) + (h-1)*width].r == 0
                           )
                        {
                            rem=true;
                        }
                        if (rem==true)
                        {
                            deleted++;
                            //cout << "deleting: " << y << " " << x << endl;
                            val.y=y;
                            val.x=x;
                            flag_for_deletion.push_back(val);
                        }
                    }
                }
            }
        }
        for (int i=0; i< flag_for_deletion.size(); i++)
        {
            input[flag_for_deletion[i].y][flag_for_deletion[i].x]=0;
        }
    }
    return deleted;
}


